//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Description of Csound environment variables.

#include <cstdlib>
#include <cstring>

#include <Message.h>
#include <Point.h>
#include <SupportDefs.h>

#include "CsAbstractJob.h"
#include "CsEnvJob.h"
#include "CsEnvMainWindow.h"

// Class name.
static const char kClassName[] = "CsoundEnvironmentJob";
static const char kJobClass[] = "Environment";

// Window title.
static const char kWindowTitle[] = "Environment Variables";

// Environment variable names.
static const char *kVariableNames[CsoundEnvironmentJob::kNumFilenames] = {
	"SFDIR",
	"SSDIR",
	"SADIR"
};

// Basic C++ object requirements: construct, copy, assign, destroy.
CsoundEnvironmentJob::CsoundEnvironmentJob(bool assignName, uint32 numFilenames)
:	inherited(assignName, numFilenames)
{
	// Inherited the program's real environment.
	for (int i = 0; i < kNumFilenames; ++i) {
		char *pVar = getenv(kVariableNames[i]);
		if (pVar) {
			SetFilename(kSoundfileDirectory + i, pVar);
		}
	}
}	

CsoundEnvironmentJob::CsoundEnvironmentJob(const CsoundEnvironmentJob &job)
:	inherited(job)
{
	// empty
}

CsoundEnvironmentJob &
CsoundEnvironmentJob::operator =(const CsoundEnvironmentJob &job)
{
	if (&job != this) {
		inherited::operator =(job);
	}
	return *this;
}

CsoundEnvironmentJob::~CsoundEnvironmentJob(void)
{
	// empty
}

CsoundAbstractJob *
CsoundEnvironmentJob::New(void) const
{
	// Virtual default constructor.
	return new CsoundEnvironmentJob();
}

CsoundAbstractJob *
CsoundEnvironmentJob::Clone(void) const
{
	// Virtual copy constructor.
	return new CsoundEnvironmentJob(*this);
}

CsoundAbstractJob &
CsoundEnvironmentJob::Become(const CsoundAbstractJob &job)
{
	// Virtual assignment operator.
	return *this = dynamic_cast<const CsoundEnvironmentJob &>(job);
}

bool
CsoundEnvironmentJob::operator ==(const CsoundAbstractJob &abstractJob) const
{
	const CsoundEnvironmentJob *const pJob =
		dynamic_cast<const CsoundEnvironmentJob *>(&abstractJob);
	if (!pJob) {
		return false;
	}
	if (pJob == this) {
		return true;
	}
	return inherited::operator ==(abstractJob);
}

//	BArchivable requirements.
CsoundEnvironmentJob::CsoundEnvironmentJob(BMessage *pMsg)
:	inherited(pMsg)
{
	// empty
}

status_t
CsoundEnvironmentJob::Archive(BMessage *pMsg, bool deep) const
{
	status_t status = inherited::Archive(pMsg, deep);

	if (status == B_NO_ERROR) {
		status = pMsg->AddString(kClassNameLabel, kClassName);
	}

	return status;
}

CsoundEnvironmentJob *
CsoundEnvironmentJob::Instantiate(BMessage *pMsg)
{
	return validate_instantiation(pMsg, kClassName)
		? new CsoundEnvironmentJob(pMsg)
		: (CsoundEnvironmentJob *)0;
}

//	CsoundAbstractJob requirements.
bool
CsoundEnvironmentJob::IsValid(void) const
{
	return true;
}

const char *
CsoundEnvironmentJob::CurrentDirectory(void) const
{
	// Dummy.
	return "";
}

char **
CsoundEnvironmentJob::CopyCommandArguments(const char *) const
{
	// Construct environment from current settings.
	const char *pDir;
	char *pEnvironment = 0;
	size_t length = 0;
	uint32 nVars = 0;
	
	for (int i = 0; i < kNumFilenames; ++i) {
		if (*(pDir = Filename(kSoundfileDirectory + i))) {
			length = AppendString(pEnvironment, kVariableNames[i], 0, length);
			length = AppendString(pEnvironment, pDir, '=', length);
			nVars += 1;
		}
	}

	// Construct environment.
	char **const pEnv = new char *[nVars + 1];		// accomodate terminating 0
	for (uint32 i = 0; i < nVars; ++i) {
		pEnv[i] = pEnvironment;
		pEnvironment += strlen(pEnvironment) + 1;
	}
	pEnv[nVars] = 0;
	
	return pEnv;
}

const char *
CsoundEnvironmentJob::JobClass(void) const
{
	return kJobClass;
}

// Construct a title for a window.
char *
CsoundEnvironmentJob::CopyWindowTitle(void) const
{
	return CopyString(kWindowTitle);
}

CsoundAbstractWindow *
CsoundEnvironmentJob::CreateWindow(BPoint origin) const
{
	// Factory method: Create an appropriate window.
	return new CsoundEnvironmentMainWindow(origin, kWindowTitle, *this);
}
