//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The environment variable window for Csound.

#include <cstring>
#include <cmath>

#include <Font.h>
#include <Point.h>
#include <Rect.h>
#include <TextControl.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsEnvJob.h"
#include "CsAbstractMainWindow.h"
#include "CsEnvMainWindow.h"

#define nElems(array) (sizeof(array) / sizeof(array[0]))

// Message fields.
static const char kFieldWhichFile[] = "CsWhich";

CsoundEnvironmentMainWindow::CsoundEnvironmentMainWindow(
	BPoint origin, const char *pTitle, const CsoundEnvironmentJob &job)
:	inherited(origin, pTitle, job)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;

	static CsoundAbstractWindow::FilenameDescriptor textControlDescriptors[] =
	{
		{	ptrCast(&CsoundEnvironmentMainWindow::mpSoundfileDirectory),
			"sfdir", "SFDIR", false,
			CsoundEnvironmentJob::kSoundfileDirectory
		},
		{	ptrCast(&CsoundEnvironmentMainWindow::mpSoundSampleDirectory),
			"ssdir", "SSDIR", false,
			CsoundEnvironmentJob::kSoundSampleDirectory
		},
		{	ptrCast(&CsoundEnvironmentMainWindow::mpSoundAnalysisDirectory),
			"sadir", "SADIR", false,
			CsoundEnvironmentJob::kSoundAnalysisDirectory
		}
	};
	const uint32 nTextControls = nElems(textControlDescriptors);
	
	CreateControls(textControlDescriptors, nTextControls, *this);
	for (i = 0; i < nTextControls; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			textControlDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.divider > maxDivider) {
			maxDivider = desc.divider;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	
	for (i = 0; i < nTextControls; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			textControlDescriptors[i];
		BTextControl *pControl = this->*(desc.pTextControl);
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, desc.height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												  pControl->TextView()->Frame().Height());
	}
	
	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	float y = spacing;
	mpSoundfileDirectory->MoveTo(spacing, y);
	mpSoundSampleDirectory->MoveTo(spacing,
											  y += textControlDescriptors[0].height);
	mpSoundAnalysisDirectory->MoveTo(spacing,
												 y += textControlDescriptors[1].height);
	pBackgroundView->AddChild(mpSoundfileDirectory);
	pBackgroundView->AddChild(mpSoundSampleDirectory);
	pBackgroundView->AddChild(mpSoundAnalysisDirectory);
	
	const float preferredWidth = maxWidth + 2 * spacing;
	const float preferredHeight =
		mpSoundAnalysisDirectory->Frame().bottom + spacing;

	// Resize to final dimensions.
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
}

CsoundEnvironmentMainWindow::~CsoundEnvironmentMainWindow(void)
{
	// empty
}

bool
CsoundEnvironmentMainWindow::QuitRequested(void)
{
	// Just hide - we are controlled by the application.
	Hide();
	return false;
}

void
CsoundEnvironmentMainWindow::SetEnvironment(const CsoundEnvironmentJob &job)
{
	// Use the values from the passed job.
	if (Lock()) {
		// ### There must be a better way to do this.
		const char *pDir;
		if (*(pDir = job.Filename(CsoundEnvironmentJob::kSoundfileDirectory))) {
			mpSoundfileDirectory->SetText(pDir);
			mpSoundfileDirectory->TextView()->SelectAll();
			mpSoundfileDirectory->Invoke();
		}
		if (*(pDir = job.Filename(CsoundEnvironmentJob::kSoundSampleDirectory))) {
			mpSoundSampleDirectory->SetText(pDir);
			mpSoundSampleDirectory->TextView()->SelectAll();
			mpSoundSampleDirectory->Invoke();
		}
		if (*(pDir = job.Filename(CsoundEnvironmentJob::kSoundAnalysisDirectory))) {
			mpSoundAnalysisDirectory->SetText(pDir);
			mpSoundAnalysisDirectory->TextView()->SelectAll();
			mpSoundAnalysisDirectory->Invoke();
		}
		Unlock();
	}
}

void
CsoundEnvironmentMainWindow::ValidJob(bool)
{
	// empty
}
