//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		A view class for monitoring output to a FILE.

#include <cerrno>
#include <csignal>
#include <cstdio>
#include <cstring>

#include <Errors.h>
#include <OS.h>
#include <Point.h>
#include <Rect.h>
#include <TextView.h>
#include <Window.h>

#include "CsTranscriptView.h"

inline BRect
calcTextRect(BRect frame)
{
	frame.OffsetTo(B_ORIGIN);
	return frame;
}

CsoundTranscriptView::CsoundTranscriptView(BRect frame, const char *pName,
														  FILE *pFile,
														  uint32 resizeMask, uint32 flags)
:	inherited(frame, pName, calcTextRect(frame),
				 resizeMask, flags|B_WILL_DRAW|B_PULSE_NEEDED|B_FRAME_EVENTS),
	mTextRect(TextRect()),
	mpFile(pFile),
	mMonitorSem(B_BAD_SEM_ID),
	mMonitorThread(B_BAD_THREAD_ID),
	mState(CsoundTranscriptView::kFinished),
	mSignaled(false),
	mTextBuffer(),
	mBufferSem(B_BAD_SEM_ID)
{
	// Initialize.  If all goes well, we will end up suspended.
	if (mpFile) {
		mBufferSem = create_sem(1, pName);
		if (mBufferSem >= B_NO_ERROR) {
			mMonitorSem = create_sem(0, pName);
			if (mMonitorSem >= B_NO_ERROR) {
				mMonitorThread = spawn_thread(MonitorFile,pName,B_NORMAL_PRIORITY,this);
				if (mMonitorThread >= B_NO_ERROR) {
					// Yo!
					mState = kSuspended;
				}
			}
		}
	}
}

CsoundTranscriptView::~CsoundTranscriptView(void)
{
	StopMonitor();
	if (mMonitorSem >= B_NO_ERROR) {
		delete_sem(mMonitorSem);
	}
	if (mBufferSem >= B_NO_ERROR) {
		delete_sem(mBufferSem);
	}
}
	
void
CsoundTranscriptView::AttachedToWindow(void)
{
	inherited::AttachedToWindow();
	StartMonitor();
}

void
CsoundTranscriptView::DetachedFromWindow(void)
{
	SuspendMonitor();
	inherited::DetachedFromWindow();
}

void
CsoundTranscriptView::FrameResized(float width, float height)
{
	// To avoid performance problems, the text rectangle is only updated rarely.
	// This happens in Pulse(), see below.
	mTextRect = BRect(0, 0, width - 1, height - 1);
}

void
CsoundTranscriptView::Pulse(void)
{
	inherited::Pulse();
	if (TextRect() != mTextRect) {
		SetTextRect(mTextRect);	
		mTextRect = TextRect();
	}
	
	// To avoid deadlocks between the monitor thread and the window thread,
	// text accumulated by the monitor thread is held in a buffer which is
	// flushed periodically.
	acquire_sem(mBufferSem);
	size_t nChars = mTextBuffer.Length();
	if (nChars) {
		Insert(TextLength(), mTextBuffer, nChars);
		mTextBuffer.Clear();
	}
	release_sem(mBufferSem);
}

int32
CsoundTranscriptView::MonitorFile(void *pData)
{
	return ((CsoundTranscriptView *)pData)->MonitorFile();
}

void
CsoundTranscriptView::SigHandler(int sig, void *pData, struct vregs *)
{
	if (sig == SIGUSR1) {
		// Record that we've been signaled. (EINTR doesn't seem to be delivered.)
		((CsoundTranscriptView *)pData)->mSignaled = true;
	}
}

int32
CsoundTranscriptView::MonitorFile(void)
{
	// The thread is controlled via signals sent by the ...Monitor() methods.
	// During normal operation, the thread waits for and processes new data.
	// If it receives a (USR1) signal, it either suspends itself or terminates,
	// depending on the value of mState.
	// The mMonitorSem is used to synchronize with the sender of the signal.
	
	status_t status = B_NO_ERROR;
	
	// Set up signal handler.
	struct sigaction oldAction, newAction, dumAction;
	newAction.sa_handler = (void (*)(int))SigHandler;
	sigemptyset(&(newAction.sa_mask));
	sigaddset(&(newAction.sa_mask), SIGUSR1);
	newAction.sa_flags = 0;
	newAction.sa_userdata = this;
	sigaction(SIGUSR1, &newAction, &oldAction);

	while (mState == kRunning) {
		char buffer[Buffer::kBufferSize];
		char *pLine = fgets(buffer, Buffer::kBufferSize, mpFile);
		if (pLine) {
			// Append text to the buffer.
			acquire_sem(mBufferSem);
			mTextBuffer += pLine;
			release_sem(mBufferSem);

		}  else if (!mSignaled) {
			// EOF or error reading the file - stop the thread right now.
			status = B_ERROR;
			mState = kFinished;
		}

		if (mSignaled) {
			// Signal received.  Must restore handler.
			mSignaled = false;
			sigaction(SIGUSR1, &newAction, &dumAction);
		
			// Suspend if requested.
			if (mState == kSuspended) {
				release_sem(mMonitorSem);
				suspend_thread(mMonitorThread);
			}
		}
	}
	
	sigaction(SIGUSR1, &oldAction, &dumAction);
	release_sem(mMonitorSem);
	return status;
}

void
CsoundTranscriptView::StartMonitor(void)
{
	if (mState == kSuspended) {
		mState = kRunning;
		if (resume_thread(mMonitorThread) < B_NO_ERROR) {
			mState = kFinished;
		}
	}
}

void
CsoundTranscriptView::SuspendMonitor(void)
{
	if (mState == kRunning) {
		mState = kSuspended;
		send_signal(mMonitorThread, SIGUSR1);
		acquire_sem(mMonitorSem);
	}
}

void
CsoundTranscriptView::StopMonitor(void)
{
	if (mState == kRunning) {
		mState = kFinished;
		send_signal(mMonitorThread, SIGUSR1);
		acquire_sem(mMonitorSem);

	} else if (mState == kSuspended) {
		mState = kFinished;
		resume_thread(mMonitorThread);
		acquire_sem(mMonitorSem);
	}
}

// Text buffer class.

CsoundTranscriptView::Buffer::Buffer(void)
:	mpBuffer(new char[CsoundTranscriptView::Buffer::kBufferSize]),
	mLength(0),
	mFree(CsoundTranscriptView::Buffer::kBufferSize)
{
	// empty
}

CsoundTranscriptView::Buffer::~Buffer(void)
{
	delete [] mpBuffer;
}

CsoundTranscriptView::Buffer &
CsoundTranscriptView::Buffer::operator +=(const char *pChars)
{
	size_t nChars = strlen(pChars);
	if (nChars > mFree) {
		char *const pNewBuffer = new char[mLength + nChars + kBufferSize];
		memcpy(pNewBuffer, mpBuffer, mLength);
		delete [] mpBuffer;
		mpBuffer = pNewBuffer;
		mFree = nChars + kBufferSize;
	}
	memcpy(mpBuffer + mLength, pChars, nChars);
	mLength += nChars;
	mFree -= nChars;
	
	return *this;
}

void
CsoundTranscriptView::Buffer::Clear(void)
{
	mFree += mLength;
	mLength = 0;
}
