//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The transcript window starts and monitors a running Csound process.

#include <cerrno>
#include <csignal>
#include <cstdio>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include <Font.h>
#include <Looper.h>
#include <Message.h>
#include <OS.h>
#include <Path.h>
#include <Point.h>
#include <Rect.h>
#include <ScrollView.h>
#include <Window.h>

#include "CsApplication.h"
#include "CsAbstractJob.h"
#include "CsTranscriptView.h"
#include "CsTranscriptWindow.h"
#include "CsEnvJob.h"
#include "CsApplication.h"

// Message fields.
static const char kFieldSource[] = "source";

// Standard input given to children.
int CsoundTranscriptWindow::smBadStdin = -1;

// Initial dimensions of transcript window in characters.
const int kTextCols = 80;
const int kTextRows = 25;
const int kTabWidth = 8;

static BRect
initialViewFrame(float &tabWidth)
{
	const float colSpacing = be_fixed_font->StringWidth(" ");
	tabWidth = kTabWidth * colSpacing;
	
	font_height fontHeight;
	be_fixed_font->GetHeight(&fontHeight);
	const float rowSpacing =
		fontHeight.ascent + fontHeight.descent + fontHeight.leading;

	BRect frame(B_ORIGIN, BPoint(kTextCols*colSpacing, kTextRows*rowSpacing));
	return frame;
}

CsoundTranscriptWindow::CsoundTranscriptWindow(
	BPoint origin, const char *pTitle, BLooper &target,
	const char *pExecutable, const CsoundAbstractJob &job)
:	inherited(BRect(origin, origin), pTitle, B_DOCUMENT_WINDOW, 0),
	mTarget(target),
	mChild(-1),
	mChildFinished(create_sem(0, pTitle))
{
	// Most of this stuff stolen from GNU Make :-)
	
	// Set up a bad standard input that reads from a broken pipe.
	if (smBadStdin == -1) {
		// Make a file descriptor that is the read end of a broken pipe.
		int pd[2];
		if (pipe(pd) == 0) {
			// Close the write side.
			(void)close(pd[1]);
			// Save the read side.
			smBadStdin = pd[0];
		}
	}
	
	// Set up a pipe to catch the child's standard output.
	int pd[2];
	if (pipe(pd) != 0) {
		// Oops.
		mTarget.PostMessage(kJobFailed);

	} else {	
		// Try to fork off a child process.
		if ((mChild = fork()) == 0) {
			// --- CHILD PROCESS --------------------------------------------------

			// Install standard file descriptors and close all other files.
			close(pd[0]);
			if (smBadStdin != 0) {
				(void)dup2(smBadStdin, 0);
			}
			if (pd[1] != 1) {
				(void)dup2(pd[1], 1);
			}
			if (pd[1] != 2) {
				(void)dup2(pd[1], 2);
			}
			int d = getdtablesize();
			while (d > 3) {
				close(--d);
			}

			// Run the program.
			chdir(job.CurrentDirectory());
			// ### BPath path = CsoundApplication::GetDirectory();
			// ### if (path.InitCheck() == B_NO_ERROR) {
			BPath path;
			if (path.SetTo(CsoundApplication::GetDirectory().Path())==B_NO_ERROR) {
				path = BPath(path.Path(), pExecutable);
			} else {
				path = BPath(pExecutable);
			}
			char **pArgv = job.CopyCommandArguments(path.Leaf());
			char **pEnv = CsoundApplication::EnvironmentJob().CopyEnvironment();
			execve(path.Path(), pArgv, pEnv);
			
			// If we come here, something went wrong.
			perror("execve");
			_exit(127);

		} else if (mChild > 0) {
			// --- PARENT PROCESS -------------------------------------------------
			
			// Get a FILE pointer for the child's output.
			close(pd[1]);
			FILE *fp = fdopen(pd[0], "r");
			
			// Create a transcript view for the file.
			float tabWidth;
			CsoundTranscriptView *const pTranscript =
				new CsoundTranscriptView(initialViewFrame(tabWidth),
												  pTitle, fp, B_FOLLOW_ALL_SIDES, 0);
			pTranscript->MakeEditable(false);
			pTranscript->SetFontAndColor(be_fixed_font);
			pTranscript->SetTabWidth(tabWidth);

			// Create a scroll view around the transcript and make it fill
			// the window.
			BScrollView *const pScroller =
				new BScrollView("scroller", pTranscript, B_FOLLOW_ALL_SIDES, 0,
									 false, true, B_NO_BORDER);
			BRect frame = pScroller->Frame();
			ResizeTo(frame.Width()-1, frame.Height());
			AddChild(pScroller);
			
			// Show the window.
			Show();
			
			// Start up a thread which will notify us when the child terminates.
			thread_id monitor =
				spawn_thread(Monitor, pTitle, B_LOW_PRIORITY, this);
			if (monitor >= B_NO_ERROR) {
				resume_thread(monitor);
			}
			// ### what to do on failure?
				
		} else {
			// --- FORK FAILED ----------------------------------------------------
			close(pd[0]);
			close(pd[1]);
			mTarget.PostMessage(kJobFailed);
		}
	}
}

CsoundTranscriptWindow::~CsoundTranscriptWindow(void)
{
	// Kill child if still running.  This shouldn't happen, but...
	KillChild(SIGKILL);

	delete_sem(mChildFinished);
}

// Hook functions.
void
CsoundTranscriptWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {
	
	// Force child process to terminate.
	case kAbortJob:
		KillChild(SIGINT);
		break;
	
	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

bool
CsoundTranscriptWindow::QuitRequested(void)
{
	KillChild(SIGINT);
	return true;
}

// Monitor the child process.
int32
CsoundTranscriptWindow::Monitor(void *pData)
{
	return ((CsoundTranscriptWindow *)pData)->Monitor();
}

int32
CsoundTranscriptWindow::Monitor(void)
{
	// Wait for the child to finish.
	pid_t pid;
	int status;
	while ((pid = wait(&status)) != mChild && !(pid < 0 && errno != EINTR))
		;

	// OK, clean up and signal the target that we're done.
	mChild = -1;
	mTarget.PostMessage(kJobFinished);
	release_sem(mChildFinished);

	return 0;
}

// Kill the child process.
void
CsoundTranscriptWindow::KillChild(int signal)
{
	if (mChild > 0) {
		kill(mChild, signal);
		acquire_sem(mChildFinished);
	}
}
