//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Abstract class describing a "job"--a collection of settings which
//		describe parameters for running Csound.
//		Each derived job class will have an associated window class, for
//		which it also acts as a factory.

#ifndef CSABSTRACTJOB_H
#define CSABSTRACTJOB_H

#include <Archivable.h>
#include <Entry.h>
#include <Message.h>
#include <SupportDefs.h>

class CsoundAbstractWindow;

class CsoundAbstractJob : public BArchivable
{
public:
	typedef BArchivable inherited;

	CsoundAbstractJob(bool assignName = true,
							uint32 numFilenames = 0,
							uint32 numNumbers = 0,
							int32 defaultOptions = 0);
	CsoundAbstractJob(const CsoundAbstractJob &job);
	CsoundAbstractJob &operator =(const CsoundAbstractJob &job);	
	virtual ~CsoundAbstractJob(void);
	
	virtual CsoundAbstractJob *New(void) const = 0;
	virtual CsoundAbstractJob *Clone(void) const = 0;
	virtual CsoundAbstractJob &Become(const CsoundAbstractJob &job) = 0;
	
	virtual bool operator ==(const CsoundAbstractJob &job) const;
	bool operator !=(const CsoundAbstractJob &job) const;
	
	// Archiving methods.
	CsoundAbstractJob(BMessage *pMsg);
	virtual status_t Archive(BMessage *pMsg, bool deep = true) const;
	static CsoundAbstractJob *Instantiate(BMessage *pMsg);

	// Saving and loading.
	void SetEntry(const entry_ref &ref);
	status_t Save(const entry_ref *pRef = (const entry_ref *)0);
	bool JobFileValid(void) const;
	
	// Filenames.
	void SetFilename(int32 which, const char *pName);
	const char *Filename(int32 which) const;
	
	// Numeric parameters.
	void SetNumber(int32 which, double value);
	double Number(int32 which) const;
	
	// Boolean and 1-of-N options.
	void ClearOptions(int32 options);				// set bits to 0
	void SetOptions(int32 options);					// set bits to 1
	void SetOptions(int32 mask, int32 value);	// replace masked bits
	int32 Options(void) const;
	
	// Check if the job is valid.
	virtual bool IsValid(void) const = 0;
	
	// Return current directory for execution.
	virtual const char *CurrentDirectory(void) const = 0;
	
	// Construct and return Csound command arguments.
	// The returned memory must be explicitly freed.
	virtual char **CopyCommandArguments(const char *pFirstArg) const = 0;

	// Return a name for a concrete derived class.
	virtual const char *JobClass(void) const = 0;

	// Construct a title for a window.
	// The returned string must be explicitly freed.
	virtual char *CopyWindowTitle(void) const = 0;
	
	// Create an instance of the concrete window class.
	virtual CsoundAbstractWindow *CreateWindow(BPoint origin) const = 0;

protected:
	// Get base name of file.
	const char *JobFilename(void) const;
	
	// String handling utilities.
	static char *CopyString(const char *const pChars);
	static size_t AppendString(char *&pChars1,
										 const char *const pChars2,
										 char separator = ' ',
										 size_t oldLength = 0);
	static size_t AppendNumber(char *&pChars1,
										 const char *const pFormat,
										 double value,
										 char separator = ' ',
										 size_t oldLength = 0);

	// Message fields for archiving.
	static const char kClassNameLabel[];
	static const char kFilenameLabel[];
	static const char kNumberLabel[];
	static const char kOptionsLabel[];

private:
	// Static data members.
	static int32 smJobCount;	// used for naming new jobs

	// Data members.
	entry_ref mRef;				// entry of corresponding file, if any
	uint32 mNumFilenames;		// number of filenames
	char **mpFilenames;			// filenames
	uint32 mNumNumbers;			// number of numeric parameters
	double *mpNumbers;			// numeric parameters
	int32 mOptions;				// boolean options
};

inline bool
CsoundAbstractJob::operator !=(const CsoundAbstractJob &job) const
{
	return !(*this == job);
}

inline const char *
CsoundAbstractJob::Filename(int32 which) const
{
	return (which < 0 || which >= mNumFilenames) ? 0 : mpFilenames[which];
}

inline double
CsoundAbstractJob::Number(int32 which) const
{
	return (which < 0 || which >= mNumNumbers) ? 0.0 : mpNumbers[which];
}

inline void
CsoundAbstractJob::ClearOptions(int32 options)
{
	mOptions &= ~options;
}

inline void
CsoundAbstractJob::SetOptions(int32 options)
{
	mOptions |= options;
}

inline void
CsoundAbstractJob::SetOptions(int32 mask, int32 value)
{
	mOptions = (mOptions & ~mask) | (value & mask);
}

inline int32
CsoundAbstractJob::Options(void) const
{
	return mOptions;
}

inline const char *
CsoundAbstractJob::JobFilename(void) const
{
	return mRef.name;
}

#endif // CSABSTRACTJOB_H
