//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Abstract class representing a 'main' window--one which has an associated
//		CsoundAbstractJob, a menu bar, and which is responsible for processing
//		control messages.  A main window may have an additional window for
//		non-essential options.

#include <cstdio>

#include <Alert.h>
#include <AppDefs.h>
#include <Bitmap.h>
#include <CheckBox.h>
#include <Directory.h>
#include <Entry.h>
#include <FilePanel.h>
#include <Handler.h>
#include <InterfaceDefs.h>
#include <List.h>
#include <Menu.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <Message.h>
#include <Messenger.h>
#include <RadioButton.h>
#include <Rect.h>
#include <TextControl.h>
#include <TextView.h>
#include <View.h>
#include <Window.h>

#include "BitmapMenuItem.h"
#include "CsApplication.h"
#include "CsAbstractJob.h"
#include "CsAbstractWindow.h"
#include "CsAbstractMainWindow.h"
#include "CsAbstractOptionsWindow.h"

// Label strings.
static const char kAppMenuLabel[] = "Csound";
static const char kAppEnvironmentLabel[] = "Environment" B_UTF8_ELLIPSIS;
static const char kAppAboutLabel[] = "About Csound" B_UTF8_ELLIPSIS;

static const char kFileMenuLabel[] = "File";
static const char kFileNewLabelFormat[] = "New %s";
static const char kFileOpenLabel[] = "Open" B_UTF8_ELLIPSIS;
static const char kFileCloseLabel[] = "Close";
static const char kFileSaveLabel[] = "Save";
static const char kFileSaveAsLabel[] = "Save As" B_UTF8_ELLIPSIS;
static const char kFileQuitLabel[] = "Quit";

static const char kPopUpBrowseLabel[] = "Browse" B_UTF8_ELLIPSIS;
static const char kPopUpEditLabel[] = "Open";

static const char kShowOptionsName[] = "shop";
static const char kShowOptionsLabel[] = "Options";

static const char kQuitWarningFormat[] =
	"Changes to '%s' have not been saved yet.";
static const char kQuitDontSaveLabel[] = "Don't Save";
static const char kQuitCancelLabel[] = "Cancel";
static const char kQuitSaveLabel[] = "Save";

static const char kSaveWarningFormat[] =
	"An error occurred while saving '%s'.";
static const char kSaveAckLabel[] = "OK";

// Message fields.
static const char kFieldDirectory[] = "directory";
static const char kFieldName[] = "name";
static const char kFieldRefs[] = "refs";
static const char kFieldSource[] = "source";

static const char kFieldFormat[] = "CsFormat";
static const char kFieldJob[] = "CsJob";
static const char kFieldMask[] = "CsMask";
static const char kFieldOption[] = "CsOption";
static const char kFieldValue[] = "CsValue";
static const char kFieldWhich[] = "CsWhich";

// Colors.
const rgb_color kViewColor = { 224, 224, 224, 0 };

// Options window title format.
static const char kOptionsWindowTitleFormat[] = "Options for %s";

// Coordinates.
const BPoint kNewWindowOffset(40, 30);

CsoundAbstractMainWindow::CsoundAbstractMainWindow(
	BPoint origin, const char *pTitle, const CsoundAbstractJob &job)
:	inherited(origin, pTitle),
	mpCurrentJob(job.Clone()),
	mpSavedJob(job.Clone()),
	mBeingModified(false),
	mpOptionsWindow((CsoundAbstractOptionsWindow *)0),
	mpShowOptions((BCheckBox *)0)
{
	BRect frame = Bounds();
	
	// Create the menu bar.
	mpMenuBar = CreateMenuBar(frame);
	AddChild(mpMenuBar);
	frame.OffsetBy(0, mpMenuBar->Frame().Height() + 1);

	// Create the window background.
	CreateBackgroundView(frame, pTitle, kViewColor);
	
	// Create the file panel.
	mpFilePanel = new BFilePanel(B_SAVE_PANEL, new BMessenger(this), NULL,
											false, false, kSaveAsJob);

	// Register with the application.
	NotifyApplication(CsoundApplication::kMainWindowCreated);
}

CsoundAbstractMainWindow::~CsoundAbstractMainWindow(void)
{
	if (mpOptionsWindow) {
		// Force options window to quit.
		mpOptionsWindow->Lock();
		mpOptionsWindow->Quit();
		mpOptionsWindow = (CsoundAbstractOptionsWindow *)0;
	}
	
	// Clean up.
	delete mpCurrentJob;
	delete mpSavedJob;
	delete mpFilePanel;
	
	// Tell the application that we're gone.
	NotifyApplication(CsoundApplication::kMainWindowDestroyed);
}

// Hook functions.
void
CsoundAbstractMainWindow::MenusBeginning(void)
{
	// Enable/disable 'Save' item depending on the state of the job.
	mpFileSaveItem->SetEnabled(
		*mpSavedJob != *mpCurrentJob && mpCurrentJob->JobFileValid());
}

void
CsoundAbstractMainWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Filing.
	case kNewJob:
		ProcessNewJob(pMsg);
		break;

	case kSaveJob:
		ProcessSaveJob(pMsg);
		break;
	
	case kSaveAsJob:
		ProcessSaveAsJob(pMsg);
		break;
	
	// Standard control messages.
	case kFilenameModified:
		ProcessFilenameModified(pMsg);
		break;

	case kSetFilename:
		ProcessSetFilename(pMsg);
		break;

	case kNumberModified:
		ProcessNumberModified(pMsg);
		break;

	case kSetNumber:
		ProcessSetNumber(pMsg);
		break;

	case kSetOption:
		ProcessSetOption(pMsg);
		break;
	
	case kSetOneOfN:
		ProcessSetOneOfN(pMsg);
		break;

	// User dragged a file (or whatever) to the window.
	case B_SIMPLE_DATA:
		ProcessDroppedData(pMsg);
		break;

	// Option window management.
	case kShowOptions:
		if (mpShowOptions && mpOptionsWindow) {
			if (mpShowOptions->Value() != B_CONTROL_OFF) {
				mpOptionsWindow->Show();
			} else {		
				mpOptionsWindow->Hide();
			}
		}
		break;
	
	case kOptionsWindowClosed:
		if (mpShowOptions) {
			mpShowOptions->SetValue(B_CONTROL_OFF);
		}
		break;
	
	// Fall back.	
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

bool
CsoundAbstractMainWindow::QuitRequested(void)
{
	// Avoid discarding unsaved changes.
	if (*mpSavedJob != *mpCurrentJob) {
		const char *pTitle = Title();
		char *pWarning = new char[strlen(pTitle) + sizeof(kQuitWarningFormat) + 1];
		sprintf(pWarning, kQuitWarningFormat, pTitle);

		BAlert *pAlert;
		if (mpCurrentJob->JobFileValid()) {
			pAlert =
				new BAlert("", pWarning,
							  kQuitDontSaveLabel, kQuitCancelLabel, kQuitSaveLabel,
							  B_WIDTH_FROM_WIDEST, B_WARNING_ALERT);
		} else {
			pAlert =
				new BAlert("", pWarning,
							  kQuitDontSaveLabel, kQuitCancelLabel, 0,
							  B_WIDTH_FROM_WIDEST, B_WARNING_ALERT);
		}
		pAlert->SetShortcut(1, B_ESCAPE);
		delete pWarning;

		switch (pAlert->Go()) {
		
		case 0:
			// Proceed.
			break;
			
		case 1:
			// Don't allow quitting.
			return false;
			
		case 2:
			// Try to save.
			if (mpCurrentJob->Save() == B_NO_ERROR) {
				mpSavedJob->Become(*mpCurrentJob);
			} else {
				WarnSavingFailed();
			}
			break;
		}
	}
	return true;
}

// Create messages targeted at the receiver.
BMessage *
CsoundAbstractMainWindow::CreateMessage(
	const CsoundAbstractWindow::ButtonDescriptor &desc) const
{
	// Descriptor specifies message code.
	BMessage *const pMsg = new BMessage(desc.msgWhat);
	return pMsg;
}

BMessage *
CsoundAbstractMainWindow::CreateMessage(
	const CsoundAbstractWindow::FilenameDescriptor &desc,
	bool modifyMessage) const
{
	// Message code is either kSetFilename or kFilenameModified.
	BMessage *const pMsg =
		new BMessage(modifyMessage ? kFilenameModified : kSetFilename);
	pMsg->AddInt32(kFieldWhich, desc.whichFile);
	return pMsg;
}

BMessage *
CsoundAbstractMainWindow::CreateMessage(
	const CsoundAbstractWindow::NumberDescriptor &desc,
	bool modifyMessage) const
{
	// Message code is either kSetNumber or kNumberModified.
	BMessage *const pMsg =
		new BMessage(modifyMessage ? kNumberModified : kSetNumber);
	pMsg->AddString(kFieldFormat, desc.pFormat);
	pMsg->AddInt32(kFieldWhich, desc.whichNumber);
	return pMsg;
}

BMessage *
CsoundAbstractMainWindow::CreateMessage(
	const CsoundAbstractWindow::CheckBoxDescriptor &desc) const
{
	// Message code is kSetOption if a positive value was specified, otherwise
	// message code is negated value.
	if (desc.option > 0) {
		BMessage *const pMsg = new BMessage(kSetOption);
		pMsg->AddInt32(kFieldOption, desc.option);
		return pMsg;
	}
	return new BMessage(-desc.option);
}

BMessage *
CsoundAbstractMainWindow::CreateMessage(
	const CsoundAbstractWindow::RadioButtonDescriptor &desc) const
{
	// Message code is always kSetOneOfN.
	BMessage *const pMsg = new BMessage(kSetOneOfN);
	pMsg->AddInt32(kFieldMask, desc.mask);
	pMsg->AddInt32(kFieldValue, desc.value);
	return pMsg;
}

// Create a title for an associated options window.
char *
CsoundAbstractMainWindow::CopyOptionsWindowTitle(void)
{
	const char *pTitle = Title();
	char *const pNewTitle =
		new char[strlen(pTitle) + sizeof(kOptionsWindowTitleFormat) + 1];
	sprintf(pNewTitle, kOptionsWindowTitleFormat, pTitle);
	return pNewTitle;
}

// Add an options window.
void
CsoundAbstractMainWindow::AddOptionsWindow(
	CsoundAbstractOptionsWindow *pOptionsWindow,
	float borderSpacing,
	BView *pRefView)
{
	mpOptionsWindow = pOptionsWindow;
	
	BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	mpShowOptions =
		new BCheckBox(emptyFrame, kShowOptionsName, kShowOptionsLabel,
							new BMessage(kShowOptions));
	mpShowOptions->SetTarget(this);
	mpShowOptions->ResizeToPreferred();
	mpShowOptions->SetValue(
		pOptionsWindow->IsHidden() ? B_CONTROL_OFF : B_CONTROL_ON);

	BView *const pBackgroundView = BackgroundView();
	BRect bgndFrame = pBackgroundView->Frame();
	
	BRect checkBoxFrame = mpShowOptions->Frame();
	const float yOffset =
		pRefView
		? (pRefView->Frame().Height() - checkBoxFrame.Height()) / 2
		: 0;
	
	const float x =
		bgndFrame.Width() - checkBoxFrame.Width() - borderSpacing;
	const float y =
		bgndFrame.Height() - checkBoxFrame.Height() - yOffset - borderSpacing;
	mpShowOptions->MoveTo(BPoint(x, y));
	pBackgroundView->AddChild(mpShowOptions);
}

// Accept a dropped file.
void
CsoundAbstractMainWindow::AcceptDroppedRef(BPoint where, const entry_ref *pRef)
{
	// Generic method for dealing with a file: If it was dropped on a text
	// control, replace its text by the filename.
	BTextControl *pControl = FindTextControl(where);
	if (pControl) {
		BPath path;
		BEntry entry(pRef);
		if (entry.GetPath(&path) == B_NO_ERROR) {
			pControl->SetText(path.Path());
			pControl->Invoke();
		}
	}
}

// Message processing.
void
CsoundAbstractMainWindow::ProcessNewJob(BMessage *pMsg)
{
	CsoundAbstractJob *pPrototype;
	if (pMsg->FindPointer(kFieldJob, &pPrototype) == B_NO_ERROR) {
		BPoint newOrigin = Frame().LeftTop() + kNewWindowOffset;
		CsoundAbstractJob *pNewJob = pPrototype->New();
		pNewJob->CreateWindow(newOrigin)->Show();
		delete pNewJob;
	}
}

void
CsoundAbstractMainWindow::ProcessSaveJob(BMessage *)
{
	if (mpCurrentJob->JobFileValid() && mpCurrentJob->Save() == B_NO_ERROR) {
		mpSavedJob->Become(*mpCurrentJob);
	} else {
		WarnSavingFailed();
	}
}
	
void
CsoundAbstractMainWindow::ProcessSaveAsJob(BMessage *pMsg)
{
	entry_ref dirRef;
	char *pName;
	if (pMsg->FindRef(kFieldDirectory, &dirRef) == B_NO_ERROR
		 && pMsg->FindString(kFieldName, &pName) == B_NO_ERROR) {
		// The message was posted by the file panel - try to save.
		BDirectory directory(&dirRef);
		BEntry entry(&directory, pName);
		entry_ref ref;
		if (entry.GetRef(&ref) == B_NO_ERROR
			 && mpCurrentJob->Save(&ref) == B_NO_ERROR) {
			mpSavedJob->Become(*mpCurrentJob);

			// Update window title.			
			char *pNewTitle = mpCurrentJob->CopyWindowTitle();
			SetTitle(pNewTitle);
			delete [] pNewTitle;

			if (mpFilePanel->IsShowing()) {
				mpFilePanel->Hide();
			}
			
		} else {
			WarnSavingFailed();
		}
	} else {
		// Show the file panel.
		mpFilePanel->Show();
	}
}

void
CsoundAbstractMainWindow::ProcessFilenameModified(BMessage *pMsg)
{
	BHandler *pHandler;
	BTextControl *pSource;
	int32 whichFile;
	if (pMsg->FindPointer(kFieldSource, &pHandler) == B_NO_ERROR
		 && (pSource = dynamic_cast<BTextControl *>(pHandler)) != 0
		 && pMsg->FindInt32(kFieldWhich, &whichFile) == B_NO_ERROR) {
		mBeingModified =
			0 != strcmp(mpCurrentJob->Filename(whichFile), pSource->Text());
	}
	ValidJob(IsJobValid());
}

void
CsoundAbstractMainWindow::ProcessSetFilename(BMessage *pMsg)
{
	BHandler *pHandler;
	BTextControl *pSource;
	int32 whichFile;
	if (pMsg->FindPointer(kFieldSource, &pHandler) == B_NO_ERROR
		 && (pSource = dynamic_cast<BTextControl *>(pHandler)) != 0
		 && pMsg->FindInt32(kFieldWhich, &whichFile) == B_NO_ERROR) {
		mpCurrentJob->SetFilename(whichFile, pSource->Text());
		mBeingModified = false;
	}
	ValidJob(IsJobValid());
}

void
CsoundAbstractMainWindow::ProcessNumberModified(BMessage *pMsg)
{
	BHandler *pHandler;
	BTextControl *pSource;
	const char *pFormat;
	int32 whichNumber;
	if (pMsg->FindPointer(kFieldSource, &pHandler) == B_NO_ERROR
		 && (pSource = dynamic_cast<BTextControl *>(pHandler)) != 0
		 && pMsg->FindString(kFieldFormat, &pFormat) == B_NO_ERROR
		 && pMsg->FindInt32(kFieldWhich, &whichNumber) == B_NO_ERROR) {
		mBeingModified =
			0 != strcmp(NumberToString(pFormat, mpCurrentJob->Number(whichNumber)),
							pSource->Text());
	}
	ValidJob(IsJobValid());
}

void
CsoundAbstractMainWindow::ProcessSetNumber(BMessage *pMsg)
{
	BHandler *pHandler;
	BTextControl *pSource;
	const char *pFormat;
	int32 whichNumber;
	if (pMsg->FindPointer(kFieldSource, &pHandler) == B_NO_ERROR
		 && (pSource = dynamic_cast<BTextControl *>(pHandler)) != 0
		 && pMsg->FindString(kFieldFormat, &pFormat) == B_NO_ERROR
		 && pMsg->FindInt32(kFieldWhich, &whichNumber) == B_NO_ERROR) {
		mpCurrentJob->SetNumber(whichNumber,
										 StringToNumber(pFormat, pSource->Text()));
		if (pSource->Window()->Lock()) {
			pSource->SetText(
				NumberToString(pFormat, mpCurrentJob->Number(whichNumber)));
			pSource->TextView()->SelectAll();
			pSource->Window()->Unlock();
		}
		mBeingModified = false;
	}
	ValidJob(IsJobValid());
}

void
CsoundAbstractMainWindow::ProcessSetOption(BMessage *pMsg)
{
	BHandler *pHandler;
	BCheckBox *pSource;
	int32 option;
	if (pMsg->FindPointer(kFieldSource, &pHandler) == B_NO_ERROR
		 && (pSource = dynamic_cast<BCheckBox *>(pHandler)) != 0
		 && pMsg->FindInt32(kFieldOption, &option) == B_NO_ERROR) {
		if (pSource->Value() != B_CONTROL_OFF) {
			mpCurrentJob->SetOptions(option);
		} else {		
			mpCurrentJob->ClearOptions(option);
		}
		ValidJob(IsJobValid());
	}
}
	
void
CsoundAbstractMainWindow::ProcessSetOneOfN(BMessage *pMsg)
{
	int32 mask, value;
	if (pMsg->FindInt32(kFieldMask, &mask) == B_NO_ERROR
		 && pMsg->FindInt32(kFieldValue, &value) == B_NO_ERROR) {
		mpCurrentJob->SetOptions(mask, value);
		ValidJob(IsJobValid());
	}
}

void
CsoundAbstractMainWindow::ProcessDroppedData(BMessage *pMsg)
{
	if (pMsg->WasDropped()) {
		BPoint dropPoint = ConvertFromScreen(pMsg->DropPoint());
		entry_ref ref;
		int32 i = 0;
		while (pMsg->FindRef(kFieldRefs, i, &ref) == B_NO_ERROR) {
			AcceptDroppedRef(dropPoint, &ref);
			++i;
		}
	}
}

// Inform the application that a main window has been created resp. destroyed.
void
CsoundAbstractMainWindow::NotifyApplication(uint32 what)
{
	BMessage msg(what);
	msg.AddPointer(CsoundApplication::kFieldWindow, this);
	be_app->PostMessage(&msg);
}

// Find a BTextControl at the given point.
BTextControl *
CsoundAbstractMainWindow::FindTextControl(BPoint where)
{
	BView *pView = FindView(where);
	while (pView) {
		BTextControl *pControl = dynamic_cast<BTextControl *>(pView);
		if (pControl) {
			return pControl;
		}
		pView = pView->Parent();
	}
	return (BTextControl *)0;
}

// Create the menu bar and the common menus.
BMenuBar *
CsoundAbstractMainWindow::CreateMenuBar(BRect frame)
{
	BMenuBar *const pMenuBar = new BMenuBar(frame, "MenuBar");

	// 'Application' menu.
	BBitmap *pAppIcon = CsoundApplication::GetIcon();
	BMenu *pAppMenu = new BMenu(kAppMenuLabel);
	BMenuItem *pAppMenuItem =
		(pAppIcon)
		? new BitmapMenuItem(pAppIcon, pAppMenu)
		: new BMenuItem(pAppMenu);
	pMenuBar->AddItem(pAppMenuItem);

	BMenuItem *pAppEnvironmentItem =
		new BMenuItem(kAppEnvironmentLabel,
						  new BMessage(CsoundApplication::kShowEnvironment),
						  'E');
	pAppEnvironmentItem->SetTarget(be_app);
	pAppMenu->AddItem(pAppEnvironmentItem);

	pAppMenu->AddSeparatorItem();

	BMenuItem *pAppAboutItem =
		new BMenuItem(kAppAboutLabel, new BMessage(B_ABOUT_REQUESTED), 'A');
	pAppAboutItem->SetTarget(be_app);
	pAppMenu->AddItem(pAppAboutItem);

	// File menu.
	BMenu *pFileMenu = new BMenu(kFileMenuLabel);
	pMenuBar->AddItem(pFileMenu);

	const BList &prototypeJobs = ((CsoundApplication *)be_app)->PrototypeJobs();
	const int32 nJobs = prototypeJobs.CountItems();
	for (int32 i = 0; i < nJobs; ++i) {
		CsoundAbstractJob *pPrototype =
			(CsoundAbstractJob *)(prototypeJobs.ItemAt(i));

		const char *pJobClass = pPrototype->JobClass();
		char *pLabel =
			new char[strlen(pJobClass) + sizeof(kFileNewLabelFormat) + 1];
		sprintf(pLabel, kFileNewLabelFormat, pJobClass);
		BMessage *pMsg = new BMessage(kNewJob);
		pMsg->AddPointer(kFieldJob, pPrototype);
		BMenuItem *pFileNewItem =
			new BMenuItem(pLabel,
							  pMsg,
							  (typeid(*pPrototype) == typeid(*mpCurrentJob)) ? 'N' : 0);
		delete pLabel;
		
		pFileNewItem->SetTarget(this);
		pFileMenu->AddItem(pFileNewItem);
	}
	
	BMenuItem *pFileOpenItem =
		new BMenuItem(kFileOpenLabel,
							new BMessage(CsoundApplication::kOpenFile),
							'O');
	pFileOpenItem->SetTarget(be_app);
	pFileMenu->AddItem(pFileOpenItem);

	BMenuItem *pFileCloseItem =
		new BMenuItem(kFileCloseLabel, new BMessage(B_QUIT_REQUESTED), 'W');
	pFileCloseItem->SetTarget(this);
	pFileMenu->AddItem(pFileCloseItem);

	pFileMenu->AddSeparatorItem();

	mpFileSaveItem =
		new BMenuItem(kFileSaveLabel, new BMessage(kSaveJob), 'S');
	mpFileSaveItem->SetTarget(this);
	pFileMenu->AddItem(mpFileSaveItem);

	BMenuItem *pFileSaveAsItem =
		new BMenuItem(kFileSaveAsLabel, new BMessage(kSaveAsJob));
	pFileSaveAsItem->SetTarget(this);
	pFileMenu->AddItem(pFileSaveAsItem);

	pFileMenu->AddSeparatorItem();

	BMenuItem *pFileQuitItem =
		new BMenuItem(kFileQuitLabel, new BMessage(B_QUIT_REQUESTED), 'Q');
	pFileQuitItem->SetTarget(be_app);
	pFileMenu->AddItem(pFileQuitItem);

	return pMenuBar;
}

// Issue a warning after a failed save.
void
CsoundAbstractMainWindow::WarnSavingFailed(void)
{
	const char *pTitle = Title();
	char *pWarning = new char[strlen(pTitle) + sizeof(kSaveWarningFormat) + 1];
	sprintf(pWarning, kSaveWarningFormat, pTitle);
	(new BAlert("", pWarning, kSaveAckLabel))->Go();
	delete pWarning;
}
