//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Abstract window class for Csound.  Serves as a base class for all 
//		the dialog-oriented windows.

#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <Button.h>
#include <CheckBox.h>
#include <Message.h>
#include <Point.h>
#include <Rect.h>
#include <RadioButton.h>
#include <TextControl.h>
#include <View.h>
#include <Window.h>

#include "CsAbstractJob.h"
#include "CsAbstractWindow.h"
#include "CsAbstractMainWindow.h"
#include "NaN.h"

const rgb_color mandatoryFilenameColor = { 200, 200, 255 };

CsoundAbstractWindow::CsoundAbstractWindow(BPoint origin, const char *pTitle)
:	inherited(BRect(origin, origin), pTitle,
				 B_TITLED_WINDOW, B_NOT_RESIZABLE|B_NOT_ZOOMABLE),
	mpBackgroundView((BView *)0)
{
	// empty
}

CsoundAbstractWindow::~CsoundAbstractWindow(void)
{
	// empty
}

// Create the window's background view.
void
CsoundAbstractWindow::CreateBackgroundView(
	BRect frame, const char *pTitle, rgb_color viewColor)
{
	mpBackgroundView = new BView(frame, pTitle, B_FOLLOW_NONE, B_WILL_DRAW);
	mpBackgroundView->SetViewColor(viewColor);
	AddChild(mpBackgroundView);
}

//	Utilities for dialog creation.
void
CsoundAbstractWindow::CreateControls(
	CsoundAbstractWindow::ButtonDescriptor descriptors[],
	uint32 nItems,
	const CsoundAbstractMainWindow &target)
{
	BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	for (uint32 i = 0; i < nItems; ++i) {
		ButtonDescriptor &desc = descriptors[i];
		BButton *pNewButton = this->*(desc.pButton) =
			new BButton(emptyFrame, desc.pName, desc.pLabel,
							 target.CreateMessage(desc));
		pNewButton->SetTarget(&target);
		pNewButton->ResizeToPreferred();
		pNewButton->GetPreferredSize(&desc.width, &desc.height);
	}
}

void
CsoundAbstractWindow::CreateControls(
	CsoundAbstractWindow::CheckBoxDescriptor descriptors[],
	uint32 nItems,
	const CsoundAbstractMainWindow &target)
{
	BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	const int32 options = target.CurrentJob().Options();
	for (uint32 i = 0; i < nItems; ++i) {
		CheckBoxDescriptor &desc = descriptors[i];
		BCheckBox *pNewCheckBox = this->*(desc.pCheckBox) =
			new BCheckBox(emptyFrame, desc.pName, desc.pLabel,
								target.CreateMessage(desc));
		pNewCheckBox->SetTarget(&target);
		pNewCheckBox->ResizeToPreferred();
		pNewCheckBox->GetPreferredSize(&desc.width, &desc.height);
		
		if (desc.option > 0) {
			const bool value = (options & desc.option) != 0;
			pNewCheckBox->SetValue(value ? B_CONTROL_ON : B_CONTROL_OFF);
		}

	}
}

void
CsoundAbstractWindow::CreateControls(
	CsoundAbstractWindow::RadioButtonDescriptor descriptors[],
	uint32 nItems,
	const CsoundAbstractMainWindow &target)
{
	BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	const int32 options = target.CurrentJob().Options();
	for (uint32 i = 0; i < nItems; ++i) {
		RadioButtonDescriptor &desc = descriptors[i];
		BRadioButton *pNewRadioButton = this->*(desc.pRadioButton) =
			new BRadioButton(emptyFrame, desc.pName, desc.pLabel,
									target.CreateMessage(desc));
		pNewRadioButton->SetTarget(&target);
		pNewRadioButton->ResizeToPreferred();
		pNewRadioButton->GetPreferredSize(&desc.width, &desc.height);

		if ((options & desc.mask) == desc.value) {
			pNewRadioButton->SetValue(B_CONTROL_ON);
		}
	}
}

void
CsoundAbstractWindow::CreateControls(
	CsoundAbstractWindow::FilenameDescriptor descriptors[],
	uint32 nItems,
	const CsoundAbstractMainWindow &target)
{
	BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	for (uint32 i = 0; i < nItems; ++i) {
		FilenameDescriptor &desc = descriptors[i];
		BTextControl *pNewTextControl = this->*(desc.pTextControl) =
			new BTextControl(emptyFrame, desc.pName, desc.pLabel,
									target.CurrentJob().Filename(desc.whichFile),
									target.CreateMessage(desc));
		pNewTextControl->SetTarget(&target);
		pNewTextControl->SetModificationMessage(target.CreateMessage(desc, true));
		pNewTextControl->ResizeToPreferred();
		pNewTextControl->GetPreferredSize(&desc.width, &desc.height);
		desc.divider = pNewTextControl->Divider();
		
		if (desc.mandatory) {
			pNewTextControl->TextView()->SetViewColor(mandatoryFilenameColor);
			pNewTextControl->TextView()->SetLowColor(mandatoryFilenameColor);
		}
	}
}

void
CsoundAbstractWindow::CreateControls(
	CsoundAbstractWindow::NumberDescriptor descriptors[],
	uint32 nItems,
	const CsoundAbstractMainWindow &target)
{
	BRect emptyFrame(B_ORIGIN, B_ORIGIN);
	for (uint32 i = 0; i < nItems; ++i) {
		NumberDescriptor &desc = descriptors[i];
		const double value = target.CurrentJob().Number(desc.whichNumber);
		BTextControl *pNewTextControl = this->*(desc.pTextControl) =
			new BTextControl(emptyFrame, desc.pName, desc.pLabel,
									NumberToString(desc.pFormat, value),
									target.CreateMessage(desc));
		pNewTextControl->SetTarget(&target);
		pNewTextControl->SetModificationMessage(target.CreateMessage(desc, true));
		pNewTextControl->ResizeToPreferred();
		pNewTextControl->GetPreferredSize(&desc.width, &desc.height);
		desc.divider = pNewTextControl->Divider();
	}
}

const char *
CsoundAbstractWindow::NumberToString(const char *pFormat, double value) const
{
	if (isNaN(value)) {
		return "";
	}
	static char buffer[100];
	sprintf(buffer, pFormat, value);
	return buffer;
}

double
CsoundAbstractWindow::StringToNumber(
	const char *pFormat, const char *pString) const
{
	if (pString[0] == '\0') {
		return aNaN();
	}
	// Arrgh.
	return atof(NumberToString(pFormat, atof(pString)));
}
