//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The heterodyne filter analysis window for Csound.

#include <cstring>
#include <cmath>

#include <Box.h>
#include <Button.h>
#include <Font.h>
#include <Menu.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <Message.h>
#include <Point.h>
#include <Rect.h>
#include <TextControl.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsHFJob.h"
#include "CsAbstractMainWindow.h"
#include "CsHFMainWindow.h"
#include "CsTranscriptWindow.h"

#define nElems(array) (sizeof(array) / sizeof(array[0]))

CsoundHFAnalysisMainWindow::CsoundHFAnalysisMainWindow(
	BPoint origin, const char *pTitle, const CsoundHFAnalysisJob &job)
:	inherited(origin, pTitle, job),
	mpTranscriptWindow((CsoundTranscriptWindow *)0)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxButtonWidth = 0;

	static CsoundAbstractWindow::FilenameDescriptor filenameDescriptors[] =
	{
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpInputFilename),
			"sinp", "Sound Input File", true,
			CsoundHFAnalysisJob::kInputFile
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpOutputFilename),
			"outp", "Analysis Output File", true,
			CsoundHFAnalysisJob::kOutputFile
		}
	};
	const uint32 nFilenames = nElems(filenameDescriptors);
	
	CreateControls(filenameDescriptors, nFilenames, *this);
	for (i = 0; i < nFilenames; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			filenameDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.divider > maxDivider) {
			maxDivider = desc.divider;
		}
	}

	static CsoundAbstractWindow::NumberDescriptor numberDescriptors[] =
	{
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpSampleRate),
			"srat", "Sample Rate", "%.0f",
			CsoundHFAnalysisJob::kSampleRate
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpChannel),
			"chnl", "Channel Number", "%.0f",
			CsoundHFAnalysisJob::kChannel
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpBeginTime),
			"btim", "Begin Time", "%.3f",
			CsoundHFAnalysisJob::kBeginTime
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpDuration),
			"durn", "Duration", "%.3f",
			CsoundHFAnalysisJob::kDuration
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpBeginFrequency),
			"bfrq", "Starting Frequency", "%.0f",
			CsoundHFAnalysisJob::kBeginFrequency
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpNumPartials),
			"nprt", "Harmonic Partials", "%.0f",
			CsoundHFAnalysisJob::kNumPartials
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpMaxAmplitude),
			"maxa", "Maximum Amplitude", "%.0f",
			CsoundHFAnalysisJob::kMaxAmplitude
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpMinAmplitude),
			"mina", "Amplitude Threshold", "%.0f",
			CsoundHFAnalysisJob::kMinAmplitude
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpNumBreakpoints),
			"nbrk", "Breakpoints", "%.0f",
			CsoundHFAnalysisJob::kNumBreakpoints
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpLowPassFilter),
			"lfrq", "Low-Pass Filter Cutoff", "%.0f",
			CsoundHFAnalysisJob::kLowPassFilter
		}
	};
	const uint32 nNumbers = nElems(numberDescriptors);
	
	CreateControls(numberDescriptors, nNumbers, *this);
	for (i = 0; i < nNumbers; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			numberDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.divider > maxDivider) {
			maxDivider = desc.divider;
		}
	}

	static CsoundAbstractWindow::ButtonDescriptor buttonDescriptors[] =
	{
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpStart),
			"strt", "Analyze",
			kStartAnalysis
		},
		{	ptrCast(&CsoundHFAnalysisMainWindow::mpAbort),
			"abrt", "Abort",	
			kAbortAnalysis
		}
	};
	const int nButtons = nElems(buttonDescriptors);
	
	CreateControls(buttonDescriptors, nButtons, *this);
	for (i = 0; i < nButtons; ++i) {
		if (buttonDescriptors[i].width > maxButtonWidth) {
			maxButtonWidth = buttonDescriptors[i].width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (nButtons * maxButtonWidth > maxWidth) {
		maxWidth = nButtons * maxButtonWidth;
	}
	
	for (i = 0; i < nFilenames; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			filenameDescriptors[i];
		BTextControl *pControl = this->*(desc.pTextControl);
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, desc.height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												  pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < nNumbers; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			numberDescriptors[i];
		BTextControl *pControl = this->*(desc.pTextControl);
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, desc.height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												  pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < nButtons; ++i) {
		const CsoundAbstractWindow::ButtonDescriptor &desc = buttonDescriptors[i];
		BButton *pControl = this->*(desc.pButton);
		pControl->ResizeTo(maxButtonWidth, desc.height);
	}
	
	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	float y = 2 * spacing;
	for (i = 0; i < nFilenames; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			filenameDescriptors[i];
		BTextControl *pTextControl = this->*(desc.pTextControl);
		pTextControl->MoveTo(spacing, y);
		y += desc.height;
	}
	BRect filesFrame(B_ORIGIN,
							BPoint(maxWidth + 2 * spacing,
									 mpOutputFilename->Frame().bottom+spacing));
	filesFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pFiles = new BBox(filesFrame, "files");
	pFiles->SetLabel("Files");
	for (i = 0; i < nFilenames; ++i) {
		pFiles->AddChild(this->*(filenameDescriptors[i].pTextControl));
	}
	pBackgroundView->AddChild(pFiles);

	y = 2 * spacing;
	for (i = 0; i < nNumbers; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			numberDescriptors[i];
		BTextControl *pTextControl = this->*(desc.pTextControl);
		pTextControl->MoveTo(spacing, y);
		y += desc.height;
	}
	BRect numbersFrame(B_ORIGIN,
							  BPoint(maxWidth + 2 * spacing,
										mpLowPassFilter->Frame().bottom+spacing));
	numbersFrame.OffsetBy(pFiles->Frame().LeftBottom());
	numbersFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pNumbers = new BBox(numbersFrame, "numbers");
	pNumbers->SetLabel("Parameters");
	for (i = 0; i < nNumbers; ++i) {
		pNumbers->AddChild(this->*(numberDescriptors[i].pTextControl));
	}
	pBackgroundView->AddChild(pNumbers);
	
	mpStart->MoveTo(spacing, pNumbers->Frame().bottom + spacing);
	mpAbort->MoveTo(mpStart->Frame().RightTop() + BPoint(spacing, 0));
	pBackgroundView->AddChild(mpStart);
	pBackgroundView->AddChild(mpAbort);
	
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = mpStart->Frame().bottom + spacing;

	// Resize to final dimensions.
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
	
	// Initialize control states.
	mpStart->MakeDefault(true);
	mpStart->SetEnabled(IsJobValid());
	mpAbort->SetEnabled(false);
}

CsoundHFAnalysisMainWindow::~CsoundHFAnalysisMainWindow(void)
{
	// empty
}

void
CsoundHFAnalysisMainWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Command button messages.
	case kStartAnalysis:
		if (!mpTranscriptWindow && IsJobValid()) {
			// Start the job in a new transcript window.
			mpTranscriptWindow =
				new CsoundTranscriptWindow(Frame().LeftTop() + BPoint(100, 100),
													 Title(), *this, "csound", CurrentJob());
			mpStart->SetEnabled(false);
			mpAbort->SetEnabled(true);			
		}
		break;
		
	case kAbortAnalysis:
		if (mpTranscriptWindow) {
			mpTranscriptWindow->PostMessage(CsoundTranscriptWindow::kAbortJob);
		}
		break;
		
	case CsoundTranscriptWindow::kJobFailed:
		if (mpTranscriptWindow && mpTranscriptWindow->Lock()) {
			mpTranscriptWindow->Quit();
		}
		(new BAlert("", "Failed to start analysis.", "Duh!"))->Go();
		// FALL THROUGH
	case CsoundTranscriptWindow::kJobFinished:
		mpTranscriptWindow = (CsoundTranscriptWindow *)0;	
		mpStart->SetEnabled(IsJobValid());
		mpAbort->SetEnabled(false);
		break;

	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

void
CsoundHFAnalysisMainWindow::ValidJob(bool valid)
{
	if (!mpTranscriptWindow) {
		mpStart->SetEnabled(valid);
	}
}
