//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The linear predictive analysis window for Csound.

#include <cstring>
#include <cmath>

#include <Box.h>
#include <Button.h>
#include <Font.h>
#include <Menu.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <Message.h>
#include <Point.h>
#include <RadioButton.h>
#include <Rect.h>
#include <TextControl.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsLPJob.h"
#include "CsAbstractMainWindow.h"
#include "CsLPMainWindow.h"
#include "CsTranscriptWindow.h"

#define nElems(array) (sizeof(array) / sizeof(array[0]))

CsoundLPAnalysisMainWindow::CsoundLPAnalysisMainWindow(
	BPoint origin, const char *pTitle, const CsoundLPAnalysisJob &job)
:	inherited(origin, pTitle, job),
	mpTranscriptWindow((CsoundTranscriptWindow *)0)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxRadioButtonWidth = 0;
	float maxButtonWidth = 0;

	static CsoundAbstractWindow::FilenameDescriptor filenameDescriptors[] =
	{
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpInputFilename),
			"sinp", "Sound Input File", true,
			CsoundLPAnalysisJob::kInputFile
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpOutputFilename),
			"outp", "Analysis Output File", true,
			CsoundLPAnalysisJob::kOutputFile
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpCommentString),
			"cmnt", "File Comment", false,
			CsoundLPAnalysisJob::kCommentString
		}
	};
	const uint32 nFilenames = nElems(filenameDescriptors);
	
	CreateControls(filenameDescriptors, nFilenames, *this);
	for (i = 0; i < nFilenames; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			filenameDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.divider > maxDivider) {
			maxDivider = desc.divider;
		}
	}

	static CsoundAbstractWindow::NumberDescriptor numberDescriptors[] =
	{
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpSampleRate),
			"srat", "Sample Rate", "%.0f",
			CsoundLPAnalysisJob::kSampleRate
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpChannel),
			"chnl", "Channel Number", "%.0f",
			CsoundLPAnalysisJob::kChannel
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpBeginTime),
			"btim", "Begin Time", "%.3f",
			CsoundLPAnalysisJob::kBeginTime
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpDuration),
			"durn", "Duration", "%.3f",
			CsoundLPAnalysisJob::kDuration
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpNumPoles),
			"npls", "Number of Poles", "%.0f",
			CsoundLPAnalysisJob::kNumPoles
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpHopSize),
			"hpsz", "Hop Size", "%.0f",
			CsoundLPAnalysisJob::kHopSize
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpMinFrequency),
			"minf", "Minimum Frequency", "%.0f",
			CsoundLPAnalysisJob::kMinFrequency
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpMaxFrequency),
			"maxf", "Maximum Frequency", "%.0f",
			CsoundLPAnalysisJob::kMaxFrequency
		}
	};
	const uint32 nNumbers = nElems(numberDescriptors);
	
	CreateControls(numberDescriptors, nNumbers, *this);
	for (i = 0; i < nNumbers; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			numberDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.divider > maxDivider) {
			maxDivider = desc.divider;
		}
	}

	static CsoundAbstractWindow::RadioButtonDescriptor radioButtonDescriptors[] =
	{
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpSilent),
			"slnt", "Silent",
			CsoundLPAnalysisJob::kVerbosityMask,
			CsoundLPAnalysisJob::kSilent
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpVerbose),
			"vrbs", "Verbose",
			CsoundLPAnalysisJob::kVerbosityMask,
			CsoundLPAnalysisJob::kVerbose
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpDebug),
			"dbug", "Debug",
			CsoundLPAnalysisJob::kVerbosityMask,
			CsoundLPAnalysisJob::kDebug
		}
	};
	const int nRadioButtons = nElems(radioButtonDescriptors);
	
	CreateControls(radioButtonDescriptors, nRadioButtons, *this);
	for (i = 0; i < nRadioButtons; ++i) {
		const CsoundAbstractWindow::RadioButtonDescriptor &desc =
			radioButtonDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.width > maxRadioButtonWidth) {
			maxRadioButtonWidth = desc.width;
		}
	}

	static CsoundAbstractWindow::ButtonDescriptor buttonDescriptors[] =
	{
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpStart),
			"strt", "Analyze",
			kStartAnalysis
		},
		{	ptrCast(&CsoundLPAnalysisMainWindow::mpAbort),
			"abrt", "Abort",	
			kAbortAnalysis
		}
	};
	const int nButtons = nElems(buttonDescriptors);
	
	CreateControls(buttonDescriptors, nButtons, *this);
	for (i = 0; i < nButtons; ++i) {
		if (buttonDescriptors[i].width > maxButtonWidth) {
			maxButtonWidth = buttonDescriptors[i].width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (nRadioButtons * maxRadioButtonWidth > maxWidth) {
		maxWidth = nRadioButtons * maxRadioButtonWidth;
	}
	if (nButtons * maxButtonWidth > maxWidth) {
		maxWidth = nButtons * maxButtonWidth;
	}
	
	for (i = 0; i < nFilenames; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			filenameDescriptors[i];
		BTextControl *pControl = this->*(desc.pTextControl);
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, desc.height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												  pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < nNumbers; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			numberDescriptors[i];
		BTextControl *pControl = this->*(desc.pTextControl);
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, desc.height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												  pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < nButtons; ++i) {
		const CsoundAbstractWindow::ButtonDescriptor &desc = buttonDescriptors[i];
		BButton *pControl = this->*(desc.pButton);
		pControl->ResizeTo(maxButtonWidth, desc.height);
	}
	
	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	float y = 2 * spacing;
	mpInputFilename->MoveTo(spacing, y);
	mpOutputFilename->MoveTo(spacing, y += filenameDescriptors[0].height);
	BRect filesFrame(B_ORIGIN,
							BPoint(maxWidth + 2 * spacing,
									 mpOutputFilename->Frame().bottom+spacing));
	filesFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pFiles = new BBox(filesFrame, "files");
	pFiles->SetLabel("Files");
	pFiles->AddChild(mpInputFilename);
	pFiles->AddChild(mpOutputFilename);
	pBackgroundView->AddChild(pFiles);

	y = 2 * spacing;
	mpCommentString->MoveTo(spacing, y);
	y += filenameDescriptors[2].height;
	for (i = 0; i < nNumbers; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			numberDescriptors[i];
		BTextControl *pTextControl = this->*(desc.pTextControl);
		pTextControl->MoveTo(spacing, y);
		y += desc.height;
	}
	BRect paramsFrame(B_ORIGIN,
							 BPoint(maxWidth + 2 * spacing,
									  mpMaxFrequency->Frame().bottom+spacing));
	paramsFrame.OffsetBy(pFiles->Frame().LeftBottom());
	paramsFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pParams = new BBox(paramsFrame, "params");
	pParams->SetLabel("Parameters");
	pParams->AddChild(mpCommentString);
	for (i = 0; i < nNumbers; ++i) {
		pParams->AddChild(this->*(numberDescriptors[i].pTextControl));
	}
	pBackgroundView->AddChild(pParams);
	
	for (i = 0; i < nRadioButtons; ++i) {
		const CsoundAbstractWindow::RadioButtonDescriptor &desc =
			radioButtonDescriptors[i];
		BRadioButton *pRadioButton = this->*(desc.pRadioButton);
		pRadioButton->MoveTo(i * maxRadioButtonWidth + spacing, 2 * spacing);
	}
	BRect messagesFrame(B_ORIGIN,
								BPoint(maxWidth + 2 * spacing,
										 mpSilent->Frame().bottom+spacing));
	messagesFrame.OffsetBy(pParams->Frame().LeftBottom());
	messagesFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pMessages = new BBox(messagesFrame, "messages");
	pMessages->SetLabel("Messages");
	for (i = 0; i < nRadioButtons; ++i) {
		pMessages->AddChild(this->*(radioButtonDescriptors[i].pRadioButton));
	}
	pBackgroundView->AddChild(pMessages);
	
	mpStart->MoveTo(spacing, pMessages->Frame().bottom + spacing);
	mpAbort->MoveTo(mpStart->Frame().RightTop() + BPoint(spacing, 0));
	pBackgroundView->AddChild(mpStart);
	pBackgroundView->AddChild(mpAbort);
	
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = mpStart->Frame().bottom + spacing;

	// Resize to final dimensions.
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
	
	// Initialize control states.
	mpStart->MakeDefault(true);
	mpStart->SetEnabled(IsJobValid());
	mpAbort->SetEnabled(false);
}

CsoundLPAnalysisMainWindow::~CsoundLPAnalysisMainWindow(void)
{
	// empty
}

void
CsoundLPAnalysisMainWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Command button messages.
	case kStartAnalysis:
		if (!mpTranscriptWindow && IsJobValid()) {
			// Start the job in a new transcript window.
			mpTranscriptWindow =
				new CsoundTranscriptWindow(Frame().LeftTop() + BPoint(100, 100),
													 Title(), *this, "csound", CurrentJob());
			mpStart->SetEnabled(false);
			mpAbort->SetEnabled(true);			
		}
		break;
		
	case kAbortAnalysis:
		if (mpTranscriptWindow) {
			mpTranscriptWindow->PostMessage(CsoundTranscriptWindow::kAbortJob);
		}
		break;
		
	case CsoundTranscriptWindow::kJobFailed:
		if (mpTranscriptWindow && mpTranscriptWindow->Lock()) {
			mpTranscriptWindow->Quit();
		}
		(new BAlert("", "Failed to start analysis.", "Duh!"))->Go();
		// FALL THROUGH
	case CsoundTranscriptWindow::kJobFinished:
		mpTranscriptWindow = (CsoundTranscriptWindow *)0;	
		mpStart->SetEnabled(IsJobValid());
		mpAbort->SetEnabled(false);
		break;

	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

void
CsoundLPAnalysisMainWindow::ValidJob(bool valid)
{
	if (!mpTranscriptWindow) {
		mpStart->SetEnabled(valid);
	}
}
