//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		Application class for Csound.

#include <Alert.h>
#include <Application.h>
#include <AppFileInfo.h>
#include <Archivable.h>
#include <Bitmap.h>
#include <Entry.h>
#include <File.h>
#include <FilePanel.h>
#include <List.h>
#include <Path.h>
#include <Roster.h>
#include <Window.h>

#include "CsApplication.h"
#include "CsAbstractJob.h"
#include "CsSynthJob.h"	
#include "CsHFJob.h"	
#include "CsLPJob.h"	
#include "CsEnvJob.h"	
#include "CsAbstractWindow.h"	
#include "CsEnvMainWindow.h"	

// Application signature.
static const char kSignature[] = "application/x-mit-csound";

// Message fields.
static const char kFieldRefs[] = "refs";
const char CsoundApplication::kFieldWindow[] = "CsWindow";

BWindow *CsoundApplication::smpEnvironmentWindow = (BWindow *)0;
BFilePanel *CsoundApplication::smpFilePanel = (BFilePanel *)0;

// BApplication requirements.
CsoundApplication::CsoundApplication(void)
:	inherited(kSignature),
	mWindowOpened(false),
	mPrototypeJobs(),
	mMainWindows()
{
	// Register prototype jobs.  Every new concrete job class must be added here.
	mPrototypeJobs.AddItem(new CsoundSynthesisJob(false));
	mPrototypeJobs.AddItem(new CsoundHFAnalysisJob(false));
	mPrototypeJobs.AddItem(new CsoundLPAnalysisJob(false));
	//mPrototypeJobs.AddItem(new CsoundPVOCAnalysisJob(false));
	//mPrototypeJobs.AddItem(new CsoundSoundinfoJob(false));
}

CsoundApplication::~CsoundApplication(void)
{
	// empty
}

// Hook functions.
void
CsoundApplication::AboutRequested(void)
{
	(new BAlert("About",
					"Csound shell for BeOS DR9\n"
					"Copyright " B_UTF8_COPYRIGHT " 1997  Jens Kilian",
					"OK"))->Go();
}

void
CsoundApplication::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Update the window list.	
	case kMainWindowCreated:
		{	BWindow *pWindow;
			if (pMsg->FindPointer(kFieldWindow, &pWindow) == B_NO_ERROR) {
				mMainWindows.AddItem(pWindow);
			}
		}
		break;
	
	case kMainWindowDestroyed:
		{	BWindow *pWindow;
			if (pMsg->FindPointer(kFieldWindow, &pWindow) == B_NO_ERROR) {
				mMainWindows.RemoveItem(pWindow);
				
				// Quit after last main window was closed.
				// The environment window doesn't count.
				if (mMainWindows.CountItems() < 2) {
					PostMessage(B_QUIT_REQUESTED);
				}
			}
		}
		break;
	
	// Open a file.  Actually, just show the file panel :-)
	case kOpenFile:
		if (smpFilePanel) {
			smpFilePanel->Show();
		}
		break;
	
	// Show the environment window.
	case kShowEnvironment:
		if (smpEnvironmentWindow) {
			if (smpEnvironmentWindow->IsHidden()) {
				smpEnvironmentWindow->Show();
			} else {
				smpEnvironmentWindow->Activate();
			}
		}
		break;
	
	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

bool
CsoundApplication::TryClosingWindow(void *pWin, void *pFlag)
{
	BWindow *pWindow = (BWindow *)pWin;
	if (pWindow->QuitRequested()) {
		if (pWindow->Lock()) {
			pWindow->Quit();
		}
	} else if (pWindow != smpEnvironmentWindow) {
		// Can't quit the application.
		*(bool *)pFlag = false;
	}
	
	return false;
}

bool
CsoundApplication::QuitRequested(void)
{
	// Try to Quit() all open main windows (except the environment window).
	bool allowed = true;
	mMainWindows.DoForEach(TryClosingWindow, &allowed);
	
	if (allowed) {
		// OK, we can also dispose of the environment window.
		smpEnvironmentWindow->Lock();
		smpEnvironmentWindow->Quit();
		smpEnvironmentWindow = (BWindow *)0;
	}

	return allowed;
}

void
CsoundApplication::ReadyToRun(void)
{
	// Create the environment window (if needed).
	SetEnvironment(CsoundEnvironmentJob(false));
	
	// Create the file panel.
	smpFilePanel = new BFilePanel;
	
	// Open an initial window if no files were loaded.
	if (!mWindowOpened) {
		CsoundSynthesisJob().CreateWindow(BPoint(100, 100))->Show();
	}
}

void
CsoundApplication::RefsReceived(BMessage *pMsg)
{
	// Loop through all the references.
	type_code type;
	int32 count;
	if (pMsg->GetInfo(kFieldRefs, &type, &count) == B_NO_ERROR) {
		for (int32 i = --count; i >= 0; i--) {
			entry_ref ref;
			if (pMsg->FindRef(kFieldRefs, i, &ref) == B_NO_ERROR) {
				BEntry entry(&ref);
				if (entry.IsFile()) {
					// Retrieve the saved job.
					BFile theFile(&ref, B_READ_ONLY);
					BMessage msg;
					if (msg.Unflatten(&theFile) == B_NO_ERROR) {
						BArchivable *pArchived = instantiate_object(&msg);
						CsoundAbstractJob *const pJob =
							dynamic_cast<CsoundAbstractJob *>(pArchived);
						if (pJob) {
							CsoundEnvironmentJob *const pEnvironment =
								dynamic_cast<CsoundEnvironmentJob *>(pJob);
							if (pEnvironment) {
								// Special case - set environment only.
								SetEnvironment(*pEnvironment);

							} else {
								// Open a window for the loaded job.
								pJob->SetEntry(ref);
								pJob->CreateWindow(BPoint(100, 100))->Show();
								mWindowOpened = true;
							}
						}
						delete pArchived;
					}
				}
			}
		}
	}
}

// Return the application's icon (0 if no icon can be found).
BBitmap *
CsoundApplication::GetIcon(void)
{
	BBitmap *pIcon = 0;

	app_info info;
	if (be_app->GetAppInfo(&info) == B_NO_ERROR) {
		BFile appFile(&(info.ref), O_RDONLY);
		BAppFileInfo appFileInfo(&appFile);
		
		pIcon = new BBitmap(BRect(0, 0, 15, 15), B_COLOR_8_BIT);
		if (appFileInfo.GetIcon(pIcon, B_MINI_ICON) != B_NO_ERROR) {
			delete pIcon;
			pIcon = 0;
		}
	}
	
	return pIcon;
}

// Return the application signature.
const char *
CsoundApplication::GetSignature(void)
{
	return kSignature;
}

// Return the application directory.
BPath
CsoundApplication::GetDirectory(void)
{
	BPath path;
	
	app_info info;
	if (be_app->GetAppInfo(&info) == B_NO_ERROR) {
		BEntry appEntry(&(info.ref));
		if (appEntry.InitCheck() == B_NO_ERROR
			 && appEntry.GetParent(&appEntry) == B_NO_ERROR) {
			appEntry.GetPath(&path);
		}
	}
	
	return path;
}

// Return a job describing environment variables.
const CsoundEnvironmentJob &
CsoundApplication::EnvironmentJob(void)
{
	const CsoundEnvironmentMainWindow &window =
		dynamic_cast<CsoundEnvironmentMainWindow &>(*smpEnvironmentWindow);
	return dynamic_cast<const CsoundEnvironmentJob &>(window.CurrentJob());
}

// Set up or modify the current environment.
void
CsoundApplication::SetEnvironment(const CsoundEnvironmentJob &job) const
{
	if (!smpEnvironmentWindow) {
		// Create the environment window.
		smpEnvironmentWindow = job.CreateWindow(BPoint(200, 200));
		smpEnvironmentWindow->Run();
	} else {
		// Modify environment.
		CsoundEnvironmentMainWindow &window =
			dynamic_cast<CsoundEnvironmentMainWindow &>(*smpEnvironmentWindow);
		window.SetEnvironment(job);
	}
}

