//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The synthesis window for Csound.  Contains the most commonly needed
//		controls.

#include <cstring>
#include <cmath>

#include <Box.h>
#include <Button.h>
#include <CheckBox.h>
#include <Font.h>
#include <Menu.h>
#include <MenuBar.h>
#include <MenuItem.h>
#include <Message.h>
#include <Mime.h>
#include <Node.h>
#include <NodeInfo.h>
#include <Point.h>
#include <Rect.h>
#include <TextControl.h>
#include <View.h>

#include "CsAbstractJob.h"
#include "CsSynthJob.h"
#include "CsAbstractMainWindow.h"
#include "CsSynthMainWindow.h"
#include "CsSynthOptionsWindow.h"
#include "CsTranscriptWindow.h"

#define nElems(array) (sizeof(array) / sizeof(array[0]))

// MIME types.
static const char kOrchestraMimeType[] = "text/x-csound-orchestra";
static const char kScoreMimeType[] = "text/x-csound-score";
static const char kTranslatedScoreMimeType[] = "text/x-csound-scoret";
static const char kMIDIMimeType[] = "audio/midi";

// Filename extensions.
static const char kOrcExtension[] = ".orc";
static const char kScoExtension[] = ".sco";
static const char kSrtExtension[] = ".srt";		// sorted score
static const char kXtrExtension[] = ".xtr";		// extract files
static const char kExtractExtension[] = ".extract";
static const char kMidiExtension[] = ".midi";	// MIDI files
static const char kMidExtension[] = ".mid";
static const char kMfExtension[] = ".mf";

// Coordinates.
const BPoint kOptionsWindowOffset(80, 60);

CsoundSynthesisMainWindow::CsoundSynthesisMainWindow(
	BPoint origin, const char *pTitle, const CsoundSynthesisJob &job)
:	inherited(origin, pTitle, job),
	mpTranscriptWindow((CsoundTranscriptWindow *)0)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxButtonWidth = 0;

	static CsoundAbstractWindow::FilenameDescriptor textControlDescriptors[] =
	{
		{	ptrCast(&CsoundSynthesisMainWindow::mpOrchestraFilename),
			"orch", "Orchestra File", true,
			CsoundSynthesisJob::kOrchestraFile
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpScoreFilename),
			"scor", "Score File", true,
			CsoundSynthesisJob::kScoreFile
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpExtractFilename),
			"extr", "Extract File", false,
			CsoundSynthesisJob::kExtractFile
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpInputFilename),
			"sinp", "Sound Input File", false,
			CsoundSynthesisJob::kInputFile
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpLineFilename),
			"linp", "Line-Oriented Input", false,
			CsoundSynthesisJob::kLineFile
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpMIDIFilename),
			"midi", "MIDI Input File", false,
			CsoundSynthesisJob::kMIDIFile
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpOutputFilename),
			"outp", "Sound Output File", false,
			CsoundSynthesisJob::kOutputFile
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpTranscriptFilename),
			"tscr", "Transcript", false,
			CsoundSynthesisJob::kTranscriptFile
		}
	};
	const uint32 nTextControls = nElems(textControlDescriptors);
	
	CreateControls(textControlDescriptors, nTextControls, *this);
	for (i = 0; i < nTextControls; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			textControlDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.divider > maxDivider) {
			maxDivider = desc.divider;
		}
	}

	static CsoundAbstractWindow::CheckBoxDescriptor checkBoxDescriptors[] =
	{
		{	ptrCast(&CsoundSynthesisMainWindow::mpTranslateScore),
			"scot", "Use SCOT Format",
			CsoundSynthesisJob::kTranslateScore
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpRealtimeMIDI),
			"rtmi", "Realtime MIDI Input",
			CsoundSynthesisJob::kRealtimeMIDI
		},
	};
	const int nCheckBoxes = nElems(checkBoxDescriptors);
	
	CreateControls(checkBoxDescriptors, nCheckBoxes, *this);
	for (i = 0; i < nCheckBoxes; ++i) {
		if (checkBoxDescriptors[i].width > maxWidth) {
			maxWidth = checkBoxDescriptors[i].width;
		}
	}
	
	static CsoundAbstractWindow::ButtonDescriptor buttonDescriptors[] =
	{
		{	ptrCast(&CsoundSynthesisMainWindow::mpStart),
			"strt", "Synthesize",
			kStartSynthesis
		},
		{	ptrCast(&CsoundSynthesisMainWindow::mpAbort),
			"abrt", "Abort",	
			kAbortSynthesis
		}
	};
	const int nButtons = nElems(buttonDescriptors);
	
	CreateControls(buttonDescriptors, nButtons, *this);
	for (i = 0; i < nButtons; ++i) {
		if (buttonDescriptors[i].width > maxButtonWidth) {
			maxButtonWidth = buttonDescriptors[i].width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (nButtons * maxButtonWidth > maxWidth) {
		maxWidth = nButtons * maxButtonWidth;
	}
	
	for (i = 0; i < nTextControls; ++i) {
		const CsoundAbstractWindow::FilenameDescriptor &desc =
			textControlDescriptors[i];
		BTextControl *pControl = this->*(desc.pTextControl);
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, desc.height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												  pControl->TextView()->Frame().Height());
	}
	for (i = 0; i < nButtons; ++i) {
		const CsoundAbstractWindow::ButtonDescriptor &desc = buttonDescriptors[i];
		BButton *pControl = this->*(desc.pButton);
		pControl->ResizeTo(maxButtonWidth, desc.height);
	}
	
	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	float y = 2 * spacing;
	mpOrchestraFilename->MoveTo(spacing, y);
	mpScoreFilename->MoveTo(spacing, y += textControlDescriptors[0].height);
	mpTranslateScore->MoveTo(spacing, y += textControlDescriptors[1].height);
	mpExtractFilename->MoveTo(spacing, y += checkBoxDescriptors[0].height);
	mpInputFilename->MoveTo(spacing, y += textControlDescriptors[2].height);
	mpLineFilename->MoveTo(spacing, y += textControlDescriptors[3].height);
	mpMIDIFilename->MoveTo(spacing, y += textControlDescriptors[4].height);
	mpRealtimeMIDI->MoveTo(spacing, y += textControlDescriptors[5].height);
	BRect inputsFrame(B_ORIGIN,
							BPoint(maxWidth + 2 * spacing,
									 mpRealtimeMIDI->Frame().bottom + spacing));
	inputsFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pInputs = new BBox(inputsFrame, "inputs");
	pInputs->SetLabel("Inputs");
	pInputs->AddChild(mpOrchestraFilename);
	pInputs->AddChild(mpScoreFilename);
	pInputs->AddChild(mpTranslateScore);
	pInputs->AddChild(mpExtractFilename);
	pInputs->AddChild(mpInputFilename);
	pInputs->AddChild(mpLineFilename);
	pInputs->AddChild(mpMIDIFilename);
	pInputs->AddChild(mpRealtimeMIDI);
	pBackgroundView->AddChild(pInputs);

	y = 2 * spacing;
	mpOutputFilename->MoveTo(spacing, y);
	mpTranscriptFilename->MoveTo(spacing, y += textControlDescriptors[6].height);
	BRect outputsFrame(B_ORIGIN,
							  BPoint(maxWidth + 2 * spacing,
										mpTranscriptFilename->Frame().bottom + spacing));
	outputsFrame.OffsetBy(pInputs->Frame().LeftBottom());
	outputsFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pOutputs = new BBox(outputsFrame, "outputs");
	pOutputs->SetLabel("Outputs");
	pOutputs->AddChild(mpOutputFilename);
	pOutputs->AddChild(mpTranscriptFilename);
	pBackgroundView->AddChild(pOutputs);

	mpStart->MoveTo(spacing, pOutputs->Frame().bottom + spacing);
	mpAbort->MoveTo(mpStart->Frame().RightTop() + BPoint(spacing, 0));
	pBackgroundView->AddChild(mpStart);
	pBackgroundView->AddChild(mpAbort);
	
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = mpStart->Frame().bottom + spacing;

	// Resize to final dimensions.
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
	
	// Initialize control states.
	mpStart->MakeDefault(true);
	mpStart->SetEnabled(IsJobValid());
	mpAbort->SetEnabled(false);

	// Create options window.
	char *const pOptionsTitle = CopyOptionsWindowTitle();
	CsoundSynthesisOptionsWindow	*pOptions =
		new CsoundSynthesisOptionsWindow(
				origin + kOptionsWindowOffset, pOptionsTitle, *this);
	delete [] pOptionsTitle;
	AddOptionsWindow(pOptions, spacing, mpAbort);
	pOptions->Run();
}

CsoundSynthesisMainWindow::~CsoundSynthesisMainWindow(void)
{
	// empty
}

void
CsoundSynthesisMainWindow::MessageReceived(BMessage *pMsg)
{
	switch (pMsg->what) {

	// Command button messages.
	case kStartSynthesis:
		if (!mpTranscriptWindow && IsJobValid()) {
			// Start the job in a new transcript window.
			mpTranscriptWindow =
				new CsoundTranscriptWindow(Frame().LeftTop() + BPoint(100, 100),
													 Title(), *this, "csound", CurrentJob());
			mpStart->SetEnabled(false);
			mpAbort->SetEnabled(true);			
		}
		break;
		
	case kAbortSynthesis:
		if (mpTranscriptWindow) {
			mpTranscriptWindow->PostMessage(CsoundTranscriptWindow::kAbortJob);
		}
		break;
		
	case CsoundTranscriptWindow::kJobFailed:
		if (mpTranscriptWindow && mpTranscriptWindow->Lock()) {
			mpTranscriptWindow->Quit();
		}
		(new BAlert("", "Failed to start synthesis.", "Duh!"))->Go();
		// FALL THROUGH
	case CsoundTranscriptWindow::kJobFinished:
		mpTranscriptWindow = (CsoundTranscriptWindow *)0;	
		mpStart->SetEnabled(IsJobValid());
		mpAbort->SetEnabled(false);
		break;

	// Fall back.
	default:
		inherited::MessageReceived(pMsg);
		break;
	}
}

void
CsoundSynthesisMainWindow::ValidJob(bool valid)
{
	if (!mpTranscriptWindow) {
		mpStart->SetEnabled(valid);
	}
}

void
CsoundSynthesisMainWindow::AcceptDroppedRef(BPoint where, const entry_ref *pRef)
{
	BPath path;
	BEntry entry(pRef);
	if (entry.GetPath(&path) == B_NO_ERROR) {
		const char *pPath = path.Path();
		BTextControl *pControl = (BTextControl *)0;
		
		// First try to map the file's mime type to one of our controls.
		static struct {
			const char *pMimeType;
			BTextControl *CsoundSynthesisMainWindow::* /*const*/ pTextControl;
		} mimeMaps[] = {
			{ kOrchestraMimeType,
			  &CsoundSynthesisMainWindow::mpOrchestraFilename
			},
			{ kScoreMimeType,
			  &CsoundSynthesisMainWindow::mpScoreFilename
			},
			{ kTranslatedScoreMimeType,
			  &CsoundSynthesisMainWindow::mpScoreFilename
			},
			{ kMIDIMimeType,
			  &CsoundSynthesisMainWindow::mpMIDIFilename
			}
		};
		const int nMimeMaps = nElems(mimeMaps);
	
		BNode node(pRef);
		BNodeInfo nodeInfo(&node);
		char mimeType[B_MIME_TYPE_LENGTH];
		if (nodeInfo.GetType(mimeType) == B_NO_ERROR) {
			for (int i = 0; !pControl && i < nMimeMaps; ++i) {
				if (strcmp(mimeType, mimeMaps[i].pMimeType) == 0) {
					pControl = this->*(mimeMaps[i].pTextControl);
				}
			}
		}
		
		// If that failed, try to map the file's extension to a control.
		// ### On success, set the mime type.
		static struct {
			const char *pExtension;
			const char *pMimeType;
			BTextControl *CsoundSynthesisMainWindow::* /*const*/ pTextControl;
		} extensionMaps[] = {
			{ kOrcExtension, kOrchestraMimeType,
			  &CsoundSynthesisMainWindow::mpOrchestraFilename
			},
			{ kScoExtension, kScoreMimeType,
			  &CsoundSynthesisMainWindow::mpScoreFilename
			},
			{ kSrtExtension, kScoreMimeType,
			  &CsoundSynthesisMainWindow::mpScoreFilename
			},
			{ kXtrExtension, (const char *)0,
			  &CsoundSynthesisMainWindow::mpExtractFilename
			},
			{ kExtractExtension, (const char *)0,
			  &CsoundSynthesisMainWindow::mpExtractFilename
			},
			{ kMidiExtension, kMIDIMimeType,
			  &CsoundSynthesisMainWindow::mpMIDIFilename
			},
			{ kMidExtension, kMIDIMimeType,
			  &CsoundSynthesisMainWindow::mpMIDIFilename
			},
			{ kMfExtension, kMIDIMimeType,
			  &CsoundSynthesisMainWindow::mpMIDIFilename
			}
		};
		const int nExtensionMaps = nElems(extensionMaps);

		const char *pExtension = strrchr(pPath, '.');
		if (pExtension) {
			for (int i = 0; !pControl && i < nExtensionMaps; ++i) {
				if (strcmp(pExtension, extensionMaps[i].pExtension) == 0) {
					pControl = this->*(extensionMaps[i].pTextControl);
					
					const char *pMimeType = extensionMaps[i].pMimeType;
					if (pMimeType) {
						nodeInfo.SetType(pMimeType);
					}
				}
			}
		}
		
		if (pControl) {
			// OK, we found a suitable control.
			pControl->SetText(path.Path());
			pControl->Invoke();
			return;
		}
	}
	
	// Fall back to inherited behavior.
	inherited::AcceptDroppedRef(where, pRef);
}
