//		Copyright © 1997	Jens Kilian
//
//		This program is free software; you can redistribute it and/or modify
//		it under the terms of the GNU General Public License as published by
//		the Free Software Foundation; either version 1, or (at your option)
//		any later version.
//
//		This program is distributed in the hope that it will be useful,
//		but WITHOUT ANY WARRANTY; without even the implied warranty of
//		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
//		GNU General Public License for more details.
//
//		You should have received a copy of the GNU General Public License
//		along with this program; if not, write to the Free Software
//		Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

//		The synthesis options window; contains less frequently used settings.

#include <cmath>

#include <Box.h>
#include <CheckBox.h>
#include <Font.h>
#include <Point.h>
#include <RadioButton.h>
#include <Rect.h>

#include "CsSynthJob.h"
#include "CsSynthMainWindow.h"
#include "CsSynthOptionsWindow.h"

#define nElems(array) (sizeof(array) / sizeof(array[0]))

CsoundSynthesisOptionsWindow::CsoundSynthesisOptionsWindow(
	BPoint origin, const char *pTitle, CsoundSynthesisMainWindow &target)
:	inherited(origin, pTitle, target)
{
	// Create all the controls and determine space requirements.
	uint32 i;
	float maxWidth = 0;
	float maxDivider = 0;
	float maxCheckBoxWidth = 0;
	float maxRadioButtonWidth = 0;

	static CsoundAbstractWindow::CheckBoxDescriptor checkBoxDescriptors[] =
	{
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpVerboseTranslation),
			"vbxl", "Verbose Orchestra Translation",
			CsoundSynthesisJob::kVerboseTranslation
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpNoteAmplitudeMsgs),
			"ampm", "Display Note Amplitude Messages",
			CsoundSynthesisJob::kNoteAmplitudeMsgs
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpOutOfRangeMsgs),
			"oorm", "Display Out-of-Range Messages",
			CsoundSynthesisJob::kOutOfRangeMsgs
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpWarningMsgs),
			"warn", "Display Warnings",
			CsoundSynthesisJob::kWarningMsgs
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpHeartbeat),
			"htbt", "Show Heartbeat",
			CsoundSynthesisJob::kHeartbeat
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpNotifyAtEndOfJob),
			"neoj", "Notify at End of Job",
			CsoundSynthesisJob::kNotifyAtEndOfJob
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpPreprocessScore),
			"iton", "Preprocess Score",
			CsoundSynthesisJob::kPreprocessScore
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpInitTimeOnly),
			"iton", "Init-Time Processing Only",
			CsoundSynthesisJob::kInitTimeOnly
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpInhibitOutput),
			"inho", "Inhibit Soundfile Output",
			CsoundSynthesisJob::kInhibitOutput
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpRewriteSoundfileHdr),
			"rsfh", "Rewrite Soundfile Header",
			CsoundSynthesisJob::kRewriteSoundfileHdr
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpDeferSoundfileLoads),
			"dsfl", "Defer Soundfile Loading",
			CsoundSynthesisJob::kDeferSoundfileLoads
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpTerminateAtEndOfTrk),
			"teot", "Terminate at End of MIDI Track",
			CsoundSynthesisJob::kTerminateAtEndOfTrk
		}
	};
	const int nCheckBoxes = nElems(checkBoxDescriptors);
	
	CreateControls(checkBoxDescriptors, nCheckBoxes, target);
	for (i = 0; i < nCheckBoxes; ++i) {
		const CsoundAbstractWindow::CheckBoxDescriptor &desc =
			checkBoxDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.width > maxCheckBoxWidth) {
			maxCheckBoxWidth = desc.width;
		}
	}

	static CsoundAbstractWindow::NumberDescriptor textControlDescriptors[] =
	{
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSoftwareSampleFrames),
			"swsf", "Software Buffer Size", "%.0f",
			CsoundSynthesisJob::kSoftwareSampleFrames
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpHardwareSampleFrames),
			"hwsf", "Hardware Buffer Size", "%.0f",
			CsoundSynthesisJob::kHardwareSampleFrames
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSampleRateOverride),
			"sovr", "Sample Rate Override", "%.0f",
			CsoundSynthesisJob::kSampleRateOverride
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpControlRateOverride),
			"kovr", "Control Rate Override", "%.0f",
			CsoundSynthesisJob::kControlRateOverride
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpTempo),
			"tmpo", "Tempo", "%.0f",
			CsoundSynthesisJob::kTempo
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSustainPedalThreshold),
			"spth", "Sustain Pedal Threshold", "%.0f",
			CsoundSynthesisJob::kSustainPedalThreshold
		}
	};
	const uint32 nTextControls = nElems(textControlDescriptors);
	
	CreateControls(textControlDescriptors, nTextControls, target);
	for (i = 0; i < nTextControls; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			textControlDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.divider > maxDivider) {
			maxDivider = desc.divider;
		}
	}

	static CsoundAbstractWindow::RadioButtonDescriptor radioButtonDescriptors[] =
	{
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSigned8BitSamples),
			"sgn8", "8-Bit Signed",
			CsoundSynthesisJob::kSampleFormatMask,
			CsoundSynthesisJob::kSigned8BitSamples
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpUnsigned8BitSamples),
			"uns8", "8-Bit Unsigned",
			CsoundSynthesisJob::kSampleFormatMask,
			CsoundSynthesisJob::kUnsigned8BitSamples
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpALawSamples),
			"alaw", "A-Law",
			CsoundSynthesisJob::kSampleFormatMask,
			CsoundSynthesisJob::kALawSamples
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpmuLawSamples),
			"ulaw", "µ-Law",
			CsoundSynthesisJob::kSampleFormatMask,
			CsoundSynthesisJob::kmuLawSamples
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSigned16BitSamples),
			"sg16", "16-Bit Signed",
			CsoundSynthesisJob::kSampleFormatMask,
			CsoundSynthesisJob::kSigned16BitSamples
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSigned32BitSamples),
			"sg32", "32-Bit Signed",
			CsoundSynthesisJob::kSampleFormatMask,
			CsoundSynthesisJob::kSigned32BitSamples
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpFloatSamples),
			"flot", "32-Bit Float",
			CsoundSynthesisJob::kSampleFormatMask,
			CsoundSynthesisJob::kFloatSamples
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpNoHeaderFormat),
			"nhdr", "Raw Data",
			CsoundSynthesisJob::kFileFormatMask,
			CsoundSynthesisJob::kNoHeaderFormat
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSoundfileFormat),
			"sndf", "Default",
			CsoundSynthesisJob::kFileFormatMask,
			CsoundSynthesisJob::kSoundfileFormat
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpAIFFFormat),
			"aiff", "AIFF",
			CsoundSynthesisJob::kFileFormatMask,
			CsoundSynthesisJob::kAIFFFormat
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpWAVFormat),
			"wave", "WAV",
			CsoundSynthesisJob::kFileFormatMask,
			CsoundSynthesisJob::kWAVFormat
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpSuppressDisplays),
			"sdpy", "None",
			CsoundSynthesisJob::kDisplayMask,
			CsoundSynthesisJob::kSuppressDisplays
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpASCIIDisplays),
			"adpy", "ASCII",
			CsoundSynthesisJob::kDisplayMask,
			CsoundSynthesisJob::kASCIIDisplays
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpPostScriptDisplays),
			"pdpy", "PostScript",
			CsoundSynthesisJob::kDisplayMask,
			CsoundSynthesisJob::kPostScriptDisplays
		},
		{	ptrCast(&CsoundSynthesisOptionsWindow::mpGraphicsDisplays),
			"gdpy", "Graphics",
			CsoundSynthesisJob::kDisplayMask,
			CsoundSynthesisJob::kGraphicsDisplays
		}
	};
	const int nRadioButtons = nElems(radioButtonDescriptors);
	
	CreateControls(radioButtonDescriptors, nRadioButtons, target);
	for (i = 0; i < nRadioButtons; ++i) {
		const CsoundAbstractWindow::RadioButtonDescriptor &desc =
			radioButtonDescriptors[i];
		if (desc.width > maxWidth) {
			maxWidth = desc.width;
		}
		if (desc.width > maxRadioButtonWidth) {
			maxRadioButtonWidth = desc.width;
		}
	}

	// Calculate overall space requirements and adjust sizes.
	font_height fontHeight;
	be_bold_font->GetHeight(&fontHeight);
	const float spacing =
		ceil((fontHeight.ascent + fontHeight.descent + fontHeight.leading) / 2);
	
	if (2 * maxCheckBoxWidth > maxWidth) {
		maxWidth = 2 * maxCheckBoxWidth;
	}
	if (maxDivider + 200 > maxWidth) {
		maxWidth = maxDivider + 200;
	}	
	if (3 * maxRadioButtonWidth + 6 * spacing > maxWidth) {
		maxWidth = 3 * maxRadioButtonWidth + 6 * spacing;
	}

	for (i = 0; i < nTextControls; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			textControlDescriptors[i];
		BTextControl *pControl = this->*(desc.pTextControl);
		// The following had to be empirically determined :-(
		pControl->ResizeTo(maxWidth, desc.height);
		pControl->SetDivider(maxDivider);
		pControl->TextView()->ResizeTo(maxWidth-maxDivider-10,
												  pControl->TextView()->Frame().Height());
	}

	// Lay out the controls.
	BView *const pBackgroundView = BackgroundView();
	
	float y = 2 * spacing;
	for (i = 0; i < (nCheckBoxes+1)/2; ++i) {
		const CsoundAbstractWindow::CheckBoxDescriptor &desc =
			checkBoxDescriptors[i];
		BCheckBox *pCheckBox = this->*(desc.pCheckBox);
		pCheckBox->MoveTo(spacing, y);
		y += desc.height;
	}
	float y1 = 2 * spacing;
	while (i < nCheckBoxes) {
		const CsoundAbstractWindow::CheckBoxDescriptor &desc =
			checkBoxDescriptors[i];
		BCheckBox *pCheckBox = this->*(desc.pCheckBox);
		pCheckBox->MoveTo(spacing + maxWidth/2, y1);
		y1 += desc.height;
		++i;
	}
	BRect switchesFrame(B_ORIGIN,
								BPoint(maxWidth + 2 * spacing, y + spacing));
	switchesFrame.OffsetBy(BPoint(spacing, spacing));
	BBox *const pSwitches = new BBox(switchesFrame, "switches");
	pSwitches->SetLabel("Switches");
	for (i = 0; i < nCheckBoxes; ++i) {
		pSwitches->AddChild(this->*(checkBoxDescriptors[i].pCheckBox));
	}
	pBackgroundView->AddChild(pSwitches);

	y = 2 * spacing;
	for (i = 0; i < nTextControls; ++i) {
		const CsoundAbstractWindow::NumberDescriptor &desc =
			textControlDescriptors[i];
		BTextControl *pTextControl = this->*(desc.pTextControl);
		pTextControl->MoveTo(spacing, y);
		y += desc.height;
	}
	BRect numbersFrame(B_ORIGIN,
							  BPoint(maxWidth + 2 * spacing,
										mpSustainPedalThreshold->Frame().bottom+spacing));
	numbersFrame.OffsetBy(pSwitches->Frame().LeftBottom());
	numbersFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pNumbers = new BBox(numbersFrame, "numbers");
	pNumbers->SetLabel("Numeric Parameters");
	for (i = 0; i < nTextControls; ++i) {
		pNumbers->AddChild(this->*(textControlDescriptors[i].pTextControl));
	}
	pBackgroundView->AddChild(pNumbers);

	y = 2 * spacing;
	for (i = 0; i < 7; ++i) {
		const CsoundAbstractWindow::RadioButtonDescriptor &desc =
			radioButtonDescriptors[i];
		BRadioButton *pRadioButton = this->*(desc.pRadioButton);
		pRadioButton->MoveTo(spacing, y);
		y += desc.height;
	}
	BRect sampleFormatsFrame(B_ORIGIN, BPoint(maxWidth/3, y + spacing));
	sampleFormatsFrame.OffsetBy(pNumbers->Frame().LeftBottom());
	sampleFormatsFrame.OffsetBy(BPoint(0, spacing));
	BBox *const pSampleFormats = new BBox(sampleFormatsFrame, "sampleFormats");
	pSampleFormats->SetLabel("Sample Formats");
	for (i = 0; i < 7; ++i) {
		pSampleFormats->AddChild(this->*(radioButtonDescriptors[i].pRadioButton));
	}
	pBackgroundView->AddChild(pSampleFormats);

	y1 = 2 * spacing;
	for (i = 7; i < 11; ++i) {
		const CsoundAbstractWindow::RadioButtonDescriptor &desc =
			radioButtonDescriptors[i];
		BRadioButton *pRadioButton = this->*(desc.pRadioButton);
		pRadioButton->MoveTo(spacing, y1);
		y1 += desc.height;
	}
	BRect fileFormatsFrame(B_ORIGIN, BPoint(maxWidth/3, y + spacing));
	fileFormatsFrame.OffsetBy(pSampleFormats->Frame().RightTop());
	fileFormatsFrame.OffsetBy(BPoint(spacing, 0));
	BBox *const pFileFormats = new BBox(fileFormatsFrame, "fileFormats");
	pFileFormats->SetLabel("File Formats");
	for (i = 7; i < 11; ++i) {
		pFileFormats->AddChild(this->*(radioButtonDescriptors[i].pRadioButton));
	}
	pBackgroundView->AddChild(pFileFormats);

	y1 = 2 * spacing;
	for (i = 11; i < nRadioButtons; ++i) {
		const CsoundAbstractWindow::RadioButtonDescriptor &desc =
			radioButtonDescriptors[i];
		BRadioButton *pRadioButton = this->*(desc.pRadioButton);
		pRadioButton->MoveTo(spacing, y1);
		y1 += desc.height;
	}
	BRect displaysFrame(B_ORIGIN, BPoint(maxWidth/3, y + spacing));
	displaysFrame.OffsetBy(pFileFormats->Frame().RightTop());
	displaysFrame.OffsetBy(BPoint(spacing, 0));
	BBox *const pDisplays = new BBox(displaysFrame, "displays");
	pDisplays->SetLabel("Displays");
	for (i = 11; i < nRadioButtons; ++i) {
		pDisplays->AddChild(this->*(radioButtonDescriptors[i].pRadioButton));
	}
	pBackgroundView->AddChild(pDisplays);

	// Resize to final dimensions.
	const float preferredWidth = maxWidth + 4 * spacing;
	const float preferredHeight = pSampleFormats->Frame().bottom + spacing;
	pBackgroundView->ResizeTo(preferredWidth, preferredHeight);
	ResizeTo(preferredWidth, pBackgroundView->Frame().bottom);
}

CsoundSynthesisOptionsWindow::~CsoundSynthesisOptionsWindow(void)
{
	// empty
}
